import os

from numpy.ma.core import MaskError

from ..cfdmimplementation import implementation
from .abstract import ReadWrite
from .exceptions import DatasetTypeError
from .netcdf import NetCDFRead


class read(ReadWrite):
    """Read field or domain constructs from a dataset.

    The following file formats are supported: netCDF and CDL.

    NetCDF files may be on local disk, on an OPeNDAP server, or in an
    S3 object store.

    The returned constructs are sorted by the netCDF variable names of
    their corresponding data or domain variables.

    **CDL files**

    A file is considered to be a CDL representation of a netCDF
    dataset if it is a text file whose first non-comment line starts
    with the seven characters "netcdf " (six letters followed by a
    space). A comment line is identified as one which starts with any
    amount white space (including none) followed by "//" (two
    slashes). It is converted to a temporary netCDF4 file using the
    external ``ncgen`` command, and the temporary file persists until
    the end of the Python session, at which time it is automatically
    deleted. The CDL file may omit data array values (as would be the
    case, for example, if the file was created with the ``-h`` or
    ``-c`` option to ``ncdump``), in which case the the relevant
    constructs in memory will be created with data with all missing
    values.

    **NetCDF unlimited dimensions**

    Domain axis constructs that correspond to NetCDF unlimited
    dimensions may be accessed with the
    `~{{package}}.DomainAxis.nc_is_unlimited` and
    `~{{package}}.DomainAxis.nc_set_unlimited` methods of a domain
    axis construct.

    **NetCDF hierarchical groups**

    Hierarchical groups in CF provide a mechanism to structure
    variables within netCDF4 datasets. Field constructs are
    constructed from grouped datasets by applying the well defined
    rules in the CF conventions for resolving references to
    out-of-group netCDF variables and dimensions. The group structure
    is preserved in the field construct's netCDF interface. Groups
    were incorporated into CF-1.8. For files with groups that state
    compliance to earlier versions of the CF conventions, the groups
    will be interpreted as per the latest release of the CF
    conventions.

    **CF-compliance**

    If the dataset is partially CF-compliant to the extent that it is
    not possible to unambiguously map an element of the netCDF dataset
    to an element of the CF data model, then a field construct is
    still returned, but may be incomplete. This is so that datasets
    which are partially conformant may nonetheless be modified in
    memory and written to new datasets.

    Such "structural" non-compliance would occur, for example, if the
    "coordinates" attribute of a CF-netCDF data variable refers to
    another variable that does not exist, or refers to a variable that
    spans a netCDF dimension that does not apply to the data
    variable. Other types of non-compliance are not checked, such
    whether or not controlled vocabularies have been adhered to. The
    structural compliance of the dataset may be checked with the
    `~{{package}}.Field.dataset_compliance` method of the returned
    constructs, as well as optionally displayed when the dataset is
    read by setting the *warnings* parameter.


    **Performance**

    Descriptive properties are always read into memory, but lazy
    loading is employed for all data arrays, unless the *to_memory*
    parameter has been set.

    .. versionadded:: (cfdm) 1.7.0

    .. seealso:: `{{package}}.write`, `{{package}}.Field`,
                 `{{package}}.Domain`, `{{package}}.unique_constructs`

    :Parameters:

        filename: `str`
            The file name or OPenDAP URL of the dataset.

            Relative paths are allowed, and standard tilde and shell
            parameter expansions are applied to the string.

            *Parameter example:*
              The file ``file.nc`` in the user's home directory could
              be described by any of the following:
              ``'$HOME/file.nc'``, ``'${HOME}/file.nc'``,
              ``'~/file.nc'``, ``'~/tmp/../file.nc'``.

        {{read external: (sequence of) `str`, optional}}

        {{read extra: (sequence of) `str`, optional}}

        {{read verbose: `int` or `str` or `None`, optional}}

        {{read warnings: `bool`, optional}}

        {{read warn_valid: `bool`, optional}}

            .. versionadded:: (cfdm) 1.8.3

        {{read mask: `bool`, optional}}

            .. versionadded:: (cfdm) 1.8.2

        {{read unpack: `bool`}}

            .. versionadded:: (cfdm) 1.11.2.0

        {{read domain: `bool`, optional}}

            .. versionadded:: (cfdm) 1.9.0.0

        {{read netcdf_backend: `None` or (sequence of) `str`, optional}}

            .. versionadded:: (cfdm) 1.11.2.0

        {{read storage_options: `dict` or `None`, optional}}

            .. versionadded:: (cfdm) 1.11.2.0

        {{read cache: `bool`, optional}}

            .. versionadded:: (cfdm) 1.11.2.0

        {{read dask_chunks: `str`, `int`, `None`, or `dict`, optional}}

              .. versionadded:: (cfdm) 1.11.2.0

        {{read store_dataset_chunks: `bool`, optional}}

            .. versionadded:: (cfdm) 1.11.2.0

        {{read cfa: `dict`, optional}}

            .. versionadded:: (cfdm) 1.12.0.0

        {{read cfa_write: (sequence of) `str`, optional}}

            .. versionadded:: (cfdm) 1.12.0.0

        {{read to_memory: (sequence of) `str`, optional}}

            .. versionadded:: (cfdm) 1.12.0.0

        {{read squeeze: `bool`, optional}}

            .. versionadded:: (cfdm) 1.12.0.0

        {{read unsqueeze: `bool`, optional}}

            .. versionadded:: (cfdm) 1.12.0.0

        {{read file_type: `None` or (sequence of) `str`, optional}}

            Valid files types are:

            ============  ============================================
            *file_type*   Description
            ============  ============================================
            ``'netCDF'``  Binary netCDF-3 or netCDF-4 file
            ``'CDL'``     Text CDL representation of a netCDF file
            ============  ============================================

            .. versionadded:: (cfdm) 1.12.0.0

        {{read ignore_unknown_type: `bool`, optional}}

            .. versionadded:: (cfdm) 1.12.0.0

    :Returns:

        `list` of `Field` or `Domain`
            The field constructs found in the dataset, or the domain
            constructs if *domain* is True. The list may be empty.

    **Examples**

    >>> x = cfdm.read('file.nc')
    >>> print(type(x))
    <type 'list'>

    Read a file and create field constructs from CF-netCDF data
    variables as well as from the netCDF variables that correspond to
    particular types metadata constructs:

    >>> f = cfdm.read('file.nc', extra='domain_ancillary')
    >>> g = cfdm.read('file.nc', extra=['dimension_coordinate',
    ...                                 'auxiliary_coordinate'])

    Read a file that contains external variables:

    >>> h = cfdm.read('parent.nc')
    >>> i = cfdm.read('parent.nc', external='external.nc')
    >>> j = cfdm.read('parent.nc', external=['external1.nc', 'external2.nc'])

    """

    implementation = implementation()

    def __new__(
        cls,
        filename,
        external=None,
        extra=None,
        verbose=None,
        warnings=False,
        warn_valid=False,
        mask=True,
        unpack=True,
        domain=False,
        netcdf_backend=None,
        storage_options=None,
        cache=True,
        dask_chunks="storage-aligned",
        store_dataset_chunks=True,
        cfa=None,
        cfa_write=None,
        to_memory=False,
        squeeze=False,
        unsqueeze=False,
        file_type=None,
        ignore_unknown_type=False,
        extra_read_vars=None,
    ):
        """Read field or domain constructs from a dataset."""
        # Initialise a netCDF read object
        netcdf = NetCDFRead(cls.implementation)
        cls.netcdf = netcdf

        filename = os.path.expanduser(os.path.expandvars(filename))

        try:
            fields = netcdf.read(
                filename,
                external=external,
                extra=extra,
                verbose=verbose,
                warnings=warnings,
                warn_valid=warn_valid,
                mask=mask,
                unpack=unpack,
                domain=domain,
                storage_options=storage_options,
                netcdf_backend=netcdf_backend,
                cache=cache,
                dask_chunks=dask_chunks,
                store_dataset_chunks=store_dataset_chunks,
                cfa=cfa,
                cfa_write=cfa_write,
                to_memory=to_memory,
                squeeze=squeeze,
                unsqueeze=unsqueeze,
                file_type=file_type,
                ignore_unknown_type=ignore_unknown_type,
                extra_read_vars=extra_read_vars,
            )
        except DatasetTypeError:
            if file_type is None:
                raise

            fields = []
        except MaskError:
            # Some data required for field interpretation is
            # missing, manifesting downstream as a NumPy
            # MaskError.
            raise ValueError(
                f"Unable to read {filename} because some netCDF "
                "variable arrays that are required for construct "
                "creation contain missing values when they shouldn't"
            )

        # Return the field or domain constructs
        return fields



from copy import deepcopy
from itertools import accumulate, product

import numpy as np
from uritools import isuri, uricompose

from ..functions import dirname
from . import abstract
from .fragment import FragmentFileArray, FragmentUniqueValueArray
from .netcdfindexer import netcdf_indexer
from .utils import chunk_locations, chunk_positions


class AggregatedArray(abstract.FileArray):
    """An array stored in a CF aggregation variable.

    .. versionadded:: (cfdm) 1.12.0.0

    """

    def __new__(cls, *args, **kwargs):
        """Store fragment array classes.

        .. versionadded:: (cfdm) 1.12.0.0

        """
        instance = super().__new__(cls)
        instance._FragmentArray = {
            "location": FragmentFileArray,
            "unique_value": FragmentUniqueValueArray,
        }
        return instance

    def __init__(
        self,
        filename=None,
        address=None,
        dtype=None,
        mask=True,
        unpack=True,
        fragment_array=None,
        attributes=None,
        storage_options=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            filename: `str`, optional
                The name of the aggregation file containing the
                aaggregation variable.

            address: `str`, optional
                The name of the aggregation variable for the array.

            dtype: `numpy.dtype`
                The data type of the aggregated data array. May be
                `None` if the numpy data-type is not known (which can
                be the case for some string types, for example).

            {{init mask: `bool`, optional}}

            {{init unpack: `bool`, optional}}

            fragment_array: `dict`
                A dictionary representation of the fragment array, in
                "location" form::

                   {'map': <'map' fragment array variable data>,
                    'location': <'location' fragment array variable data>,
                    'variable': <'variable' fragment array variable data>,}

                or "unique_value" form:

                   {'map': <'map' fragment array variable data>,
                    'unique_value': <'unique_value' fragment array data>}

            storage_options: `dict` or `None`, optional
                Key/value pairs to be passed on to the creation of
                `s3fs.S3FileSystem` file systems to control the
                opening of fragment files in S3 object stores. Ignored
                for files not in an S3 object store, i.e. those whose
                names do not start with ``s3:``.

                By default, or if `None`, then *storage_options* is
                taken as ``{}``.

                If the ``'endpoint_url'`` key is not in
                *storage_options* or is not in a dictionary defined by
                the ``'client_kwargs`` key (which is always the case
                when *storage_options* is `None`), then one will be
                automatically inserted for accessing a fragment S3
                file. For example, for a file name of
                ``'s3://store/data/file.nc'``, an ``'endpoint_url'``
                key with value ``'https://store'`` would be created.

                *Parameter example:*
                  ``{'key: 'scaleway-api-key...', 'secret':
                  'scaleway-secretkey...', 'endpoint_url':
                  'https://s3.fr-par.scw.cloud', 'client_kwargs':
                  {'region_name': 'fr-par'}}``

            {{init attributes: `dict` or `None`, optional}}

                If *attributes* is `None`, the default, then the
                attributes will be set from the netCDF variable during
                the first `__getitem__` call.

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            filename=filename,
            address=address,
            dtype=dtype,
            mask=True,
            unpack=unpack,
            attributes=attributes,
            storage_options=storage_options,
            source=source,
            copy=copy,
        )

        if source is not None:
            try:
                shape = source.shape
            except AttributeError:
                shape = None

            try:
                fragment_array_shape = source.get_fragment_array_shape()
            except AttributeError:
                fragment_array_shape = None

            try:
                fragment_array = source.get_fragment_array(copy=False)
            except AttributeError:
                fragment_array = {}

            try:
                fragment_type = source.get_fragment_type()
            except AttributeError:
                fragment_type = None
        else:
            if filename is not None:
                (
                    shape,
                    fragment_array_shape,
                    fragment_type,
                    fragment_array,
                ) = self._parse_fragment_array(filename, fragment_array)
            else:
                shape = None
                fragment_array_shape = None
                fragment_array = None
                fragment_type = None

        self._set_component("shape", shape, copy=False)
        self._set_component(
            "fragment_array_shape", fragment_array_shape, copy=False
        )
        self._set_component("fragment_array", fragment_array, copy=False)
        self._set_component("fragment_type", fragment_type, copy=False)

    def __getitem__(self, index):
        """Return a subspace.

        .. versionadded:: (cfdm) 1.12.0.0

        """
        dx = netcdf_indexer(
            self.to_dask_array(),
            mask=True,
            unpack=False,
            always_masked_array=False,
            orthogonal_indexing=True,
            attributes=self.get_attributes(),
            copy=False,
        )
        return dx[index].compute()

    @property
    def __in_memory__(self):
        """True if the array data is in memory.

        .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `bool`

        """
        return False

    def _parse_fragment_array(self, aggregated_filename, fragment_array):
        """Parse the fragment array dictionary.

        .. versionadded:: (cfdm) 1.12.0.0

        :Parameters:

            aggregated_filename: `str`
                The name of the aggregation file.

            fragment_array: `dict`
               A dictionary representation of the fragment array, in
               "location" form::

                  {'map': <'map' fragment array variable data>,
                   'location': <'location' fragment array variable data>,
                   'variable': <'variable' fragment array variable data>}

               or "unique_value" form::

                  {'map': <'map' fragment array variable data>,
                   'unique_value': <'unique_value' fragment array data>}

        :Returns:

            4-`tuple`
                1. The shape of the aggregated data.
                2. The shape of the array of fragments.
                3. The type of the fragments (either
                   ``'unique_value'`` or ``'location'``).
                4. The parsed aggregation instructions.

        """
        parsed_fragment_array = {}

        fa_map = fragment_array["map"]
        if fa_map.ndim:
            compressed = np.ma.compressed
            chunks = [compressed(i).tolist() for i in fa_map]
        else:
            # Scalar 'map' variable
            chunks = []

        aggregated_shape = tuple([sum(c) for c in chunks])
        fragment_array_indices = chunk_positions(chunks)
        fragment_shapes = chunk_locations(chunks)

        if "location" in fragment_array:
            # --------------------------------------------------------
            # Each fragment is in a file, rather than given by a
            # unique value.
            # --------------------------------------------------------
            fragment_type = "location"
            fa_variable = fragment_array["variable"]
            fa_location = fragment_array["location"]
            fragment_array_shape = fa_location.shape

            if not fa_variable.ndim:
                fa_variable = fa_variable.item()
                scalar = True
            else:
                scalar = False

            for index, shape in zip(fragment_array_indices, fragment_shapes):
                if scalar:
                    variable = fa_variable
                else:
                    variable = fa_variable[index].item()

                parsed_fragment_array[index] = {
                    "map": shape,
                    "location": fa_location[index].item(),
                    "variable": variable,
                }
        else:
            # --------------------------------------------------------
            # Each fragment comprises a unique value, rather than
            # being in a file.
            # --------------------------------------------------------
            fragment_type = "unique_value"
            fa_unique_value = fragment_array["unique_value"]
            fragment_array_shape = fa_unique_value.shape
            parsed_fragment_array = {
                index: {
                    "map": shape,
                    "unique_value": fa_unique_value[index].item(),
                }
                for index, shape in zip(
                    fragment_array_indices, fragment_shapes
                )
            }

        return (
            aggregated_shape,
            fragment_array_shape,
            fragment_type,
            parsed_fragment_array,
        )

    def get_fragment_array(self, copy=True):
        """Get the aggregation data dictionary.

        The aggregation data dictionary contains the definitions of
        the fragments and the instructions on how to aggregate them.
        The keys are indices of the fragment array dimensions,
        e.g. ``(1, 0, 0, 0)``.

        .. versionadded:: (cfdm) 1.12.0.0

         .. seealso:: `get_fragment_type`,
                      `get_fragment_array_shape`,
                      `get_fragmented_dimensions`

        :Parameters:

            copy: `bool`, optional
                Whether or not to return a copy of the aggregation
                dictionary. By default a deep copy is returned.

                .. warning:: If False then changing the returned
                             dictionary in-place will change the
                             aggregation dictionary stored in the
                             {{class}} instance, **as well as in any
                             copies of it**.

        :Returns:

            `dict`
                The aggregation data dictionary.

        **Examples**

        >>> a.shape
        (12, 1, 73, 144)
        >>> a.get_fragment_array_shape()
        (2, 1, 1, 1)
        >>> a.get_fragment_array()
        {(0, 0, 0, 0): {
          'file': ('January-June.nc',),
          'variable': ('temp',),
          'format': 'nc',
          'location': [(0, 6), (0, 1), (0, 73), (0, 144)]},
         (1, 0, 0, 0): {
          'file': ('July-December.nc',),
          'variable': ('temp',),
          'format': 'nc',
          'location': [(6, 12), (0, 1), (0, 73), (0, 144)]}}

        """
        fragment_array = self._get_component("fragment_array")
        if copy:
            fragment_array = deepcopy(fragment_array)

        return fragment_array

    def get_fragment_array_shape(self):
        """Get the sizes of the fragment dimensions.

        The fragment dimension sizes are given in the same order as
        the aggregated dimension sizes given by `shape`.

        .. versionadded:: (cfdm) 1.12.0.0

         .. seealso:: `get_fragment_array`,
                      `get_fragment_type`,
                      `get_fragmented_dimensions`

        :Returns:

            `tuple`
                The shape of the fragment dimensions.

        """
        return self._get_component("fragment_array_shape")

    def get_fragment_type(self):
        """The type of fragments in the fragment array.

        Either ``'location'`` to indicate that the fragments are
        files, or else ``'unique_value'`` to indicate that they
        are represented by their unique data values.

        .. versionadded:: (cfdm) 1.12.0.0

         .. seealso:: `get_fragment_array`,
                      `get_fragment_array_shape`,
                      `get_fragmented_dimensions`

        :Returns:

            `str`
                The fragment type.

        """
        return self._get_component("fragment_type", None)

    def get_fragmented_dimensions(self):
        """The positions of dimensions spanned by two or more fragments.

        .. versionadded:: (cfdm) 1.12.0.0

         .. seealso:: `get_fragment_array`,
                      `get_fragment_array_shape`,
                      `get_fragment_type`

        :Returns:

            `list`
                The dimension positions.

        **Examples**

        >>> a.get_fragment_array_shape()
        (20, 1, 40, 1)
        >>> a.get_fragmented_dimensions()
        [0, 2]

        >>> a.get_fragment_array_shape()
        (1, 1, 1)
        >>> a.get_fragmented_dimensions()
        []

        """
        return [
            i
            for i, size in enumerate(self.get_fragment_array_shape())
            if size > 1
        ]

    def subarray_shapes(self, shapes):
        """Create the subarray shapes.

        A fragmented dimenion (i.e. one spanned by two or fragments)
        will always have a subarray size equal to the size of each of
        its fragments, overriding any other size implied by the
        *shapes* parameter.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `subarrays`

        :Parameters:

            shapes: `int`, sequence, `dict` or `str`, optional
                Define the subarray shapes.

                Any value accepted by the *chunks* parameter of the
                `dask.array.from_array` function is allowed.

                The subarray sizes implied by *chunks* for a dimension
                that has been fragmented are ignored, so their
                specification is arbitrary.

        :Returns:

            `tuple`
                The subarray sizes along each dimension.

        **Examples**

        >>> a.shape
        (12, 1, 73, 144)
        >>> a.get_fragment_array_shape()
        (2, 1, 1, 1)
        >>> a.fragmented_dimensions()
        [0]
        >>> a.subarray_shapes(-1)
        ((6, 6), (1,), (73,), (144,))
        >>> a.subarray_shapes(None)
        ((6, 6), (1,), (73,), (144,))
        >>> a.subarray_shapes("auto")
        ((6, 6), (1,), (73,), (144,))
        >>> a.subarray_shapes((None, 1, 40, 50))
        ((6, 6), (1,), (40, 33), (50, 50, 44))
        >>>  a.subarray_shapes((None, None, "auto", 50))
        ((6, 6), (1,), (73,), (50, 50, 44))
        >>>  a.subarray_shapes({2: 40})
        ((6, 6), (1,), (40, 33), (144,))

        """
        from numbers import Number

        from dask.array.core import normalize_chunks

        # Positions of fragmented dimensions (i.e. those spanned by
        # two or more fragments)
        f_dims = self.get_fragmented_dimensions()

        shape = self.shape
        fragment_array = self.get_fragment_array(copy=False)

        # Create the base chunks.
        chunks = []
        ndim = self.ndim
        for dim, (n_fragments, size) in enumerate(
            zip(self.get_fragment_array_shape(), self.shape)
        ):
            if dim in f_dims:
                # This aggregated dimension is spanned by two or more
                # fragments => set the chunks to be the same size as
                # the each fragment.
                c = []
                index = [0] * ndim
                for j in range(n_fragments):
                    index[dim] = j
                    loc = fragment_array[tuple(index)]["map"][dim]
                    chunk_size = loc[1] - loc[0]
                    c.append(chunk_size)

                chunks.append(tuple(c))
            else:
                # This aggregated dimension is spanned by exactly one
                # fragment => store `None` for now. This will get
                # overwritten from 'shapes'.
                chunks.append(None)

        if isinstance(shapes, (str, Number)) or shapes is None:
            chunks = [
                c if i in f_dims else shapes for i, c in enumerate(chunks)
            ]
        elif isinstance(shapes, dict):
            chunks = [
                chunks[i] if i in f_dims else shapes.get(i, "auto")
                for i, c in enumerate(chunks)
            ]
        else:
            # chunks is a sequence
            if len(shapes) != ndim:
                raise ValueError(
                    f"Wrong number of 'shapes' elements in {shapes}: "
                    f"Got {len(shapes)}, expected {self.ndim}"
                )

            chunks = [
                c if i in f_dims else shapes[i] for i, c in enumerate(chunks)
            ]

        return normalize_chunks(chunks, shape=shape, dtype=self.dtype)

    def subarrays(self, subarray_shapes):
        """Return descriptors for every subarray.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `subarray_shapes`

        :Parameters:

            subarray_shapes: `tuple`
                The subarray sizes along each dimension, as returned
                by a prior call to `subarray_shapes`.

        :Returns:

            6-`tuple` of iterators
               Each iterator iterates over a particular descriptor
               from each subarray.

               1. The indices of the aggregated array that correspond
                  to each subarray.

               2. The shape of each subarray.

               3. The indices of the fragment that corresponds to each
                  subarray (some subarrays may be represented by a
                  part of a fragment).

               4. The location of each subarray.

               5. The location on the fragment dimensions of the
                  fragment that corresponds to each subarray.

               6. The shape of each fragment that overlaps each chunk.

        **Examples**

        An aggregated array with shape (12, 73, 144) has two
        fragments, both with with shape (6, 73, 144).

        >>> a.shape
        (12, 73, 144)
        >>> a.get_fragment_array_shape()
        (2, 1, 1)
        >>> a.fragmented_dimensions()
        [0]
        >>> subarray_shapes = a.subarray_shapes({1: 40})
        >>> print(subarray_shapes)
        ((6, 6), (40, 33), (144,))
        >>> (
        ...  u_indices,
        ...  u_shapes,
        ...  f_indices,
        ...  s_locations,
        ...  f_locations,
        ...  f_shapes,
        ... ) = a.subarrays(subarray_shapes)
        >>> for i in u_indices:
        ...    print(i)
        ...
        (slice(0, 6, None), slice(0, 40, None), slice(0, 144, None))
        (slice(0, 6, None), slice(40, 73, None), slice(0, 144, None))
        (slice(6, 12, None), slice(0, 40, None), slice(0, 144, None))
        (slice(6, 12, None), slice(40, 73, None), slice(0, 144, None))

        >>> for i in u_shapes
        ...    print(i)
        ...
        (6, 40, 144)
        (6, 33, 144)
        (6, 40, 144)
        (6, 33, 144)
        >>> for i in f_indices:
        ...    print(i)
        ...
        (slice(None, None, None), slice(0, 40, None), slice(0, 144, None))
        (slice(None, None, None), slice(40, 73, None), slice(0, 144, None))
        (slice(None, None, None), slice(0, 40, None), slice(0, 144, None))
        (slice(None, None, None), slice(40, 73, None), slice(0, 144, None))
        >>> for i in s_locations:
        ...    print(i)
        ...
        (0, 0, 0)
        (0, 1, 0)
        (1, 0, 0)
        (1, 1, 0)
        >>> for i in f_locations:
        ...    print(i)
        ...
        (0, 0, 0)
        (0, 0, 0)
        (1, 0, 0)
        (1, 0, 0)
        >>> for i in f_shapes:
        ...    print(i)
        ...
        (6, 73, 144)
        (6, 73, 144)
        (6, 73, 144)
        (6, 73, 144)

        """
        f_dims = self.get_fragmented_dimensions()

        # The indices of the uncompressed array that correspond to
        # each subarray, the shape of each uncompressed subarray, and
        # the location of each subarray
        s_locations = []
        u_shapes = []
        u_indices = []
        f_locations = []
        for dim, c in enumerate(subarray_shapes):
            nc = len(c)
            s_locations.append(tuple(range(nc)))
            u_shapes.append(c)

            if dim in f_dims:
                f_locations.append(tuple(range(nc)))
            else:
                # No fragmentation along this dimension
                f_locations.append((0,) * nc)

            c = tuple(accumulate((0,) + c))
            u_indices.append([slice(i, j) for i, j in zip(c[:-1], c[1:])])

        # For each subarray, the part of the fragment that corresponds
        # to it.
        f_indices = [
            (slice(None),) * len(u) if dim in f_dims else u
            for dim, u in enumerate(u_indices)
        ]

        # For each subarray, the shape of the fragment that
        # corresponds to it.
        f_shapes = [
            u_shape if dim in f_dims else (size,) * len(u_shape)
            for dim, (u_shape, size) in enumerate(zip(u_shapes, self.shape))
        ]

        return (
            product(*u_indices),
            product(*u_shapes),
            product(*f_indices),
            product(*s_locations),
            product(*f_locations),
            product(*f_shapes),
        )

    def to_dask_array(self, chunks="auto"):
        """Create a dask array with `FragmentArray` chunks.

        .. versionadded:: (cfdm) 1.12.0.0

        :Parameters:

            chunks: `int`, `tuple`, `dict` or `str`, optional
                Specify the chunking of the returned dask array.

                Any value accepted by the *chunks* parameter of the
                `dask.array.from_array` function is allowed.

                The chunk sizes implied by *chunks* for a dimension that
                has been fragmented are ignored and replaced with values
                that are implied by that dimensions fragment sizes.

        :Returns:

            `dask.array.Array`

        """
        import dask.array as da
        from dask.array.core import getter
        from dask.base import tokenize

        name = (f"{self.__class__.__name__}-{tokenize(self)}",)

        dtype = self.dtype
        fragment_array = self.get_fragment_array(copy=False)
        storage_options = self.get_storage_options()
        fragment_type = self.get_fragment_type()
        aggregated_attributes = self.get_attributes()
        unpack = self.get_unpack()

        if fragment_type == "location":
            # Get the directory of the aggregation file as an absolute
            # URI
            aggregation_file_directory = dirname(self.get_filename())
            if not isuri(aggregation_file_directory):
                aggregation_file_directory = uricompose(
                    scheme="file",
                    authority="",
                    path=aggregation_file_directory,
                )

        # Set the chunk sizes for the dask array
        chunks = self.subarray_shapes(chunks)

        try:
            FragmentArray = self._FragmentArray[fragment_type]
        except KeyError:
            raise ValueError(
                "Can't get fragment array class for unknown "
                f"fragment type: {fragment_type!r}"
            )

        dsk = {}
        for (
            u_indices,
            u_shape,
            f_indices,
            chunk_index,
            fragment_index,
            fragment_shape,
        ) in zip(*self.subarrays(chunks)):
            kwargs = fragment_array[fragment_index].copy()
            kwargs.pop("map", None)

            if fragment_type == "location":
                kwargs["filename"] = kwargs.pop("location")
                kwargs["address"] = kwargs.pop("variable")
                kwargs["storage_options"] = storage_options
                kwargs["aggregation_file_directory"] = (
                    aggregation_file_directory
                )

            fragment = FragmentArray(
                dtype=dtype,
                shape=fragment_shape,
                unpack_aggregated_data=unpack,
                aggregated_attributes=aggregated_attributes,
                **kwargs,
            )

            key = f"{fragment.__class__.__name__}-{tokenize(fragment)}"
            dsk[key] = fragment
            dsk[name + chunk_index] = (
                getter,
                key,
                f_indices,
                False,
                False,
            )

        # Return the dask array
        return da.Array(dsk, name[0], chunks=chunks, dtype=dtype)



from ..numpyarray import NumpyArray


class DeprecationError(Exception):
    """Deprecation error."""

    pass


class NetCDFFileMixin:
    """Mixin class for netCDF file arrays.

    .. versionadded:: (cfdm) 1.11.2.0

    """

    def _group(self, dataset, groups):
        """Return the group object containing a variable.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            dataset: `netCDF4.Dataset` or `h5netcdf.File`
                The dataset containing the variable.

            groups: sequence of `str`
                The definition of which group the variable is in. For
                instance, of the variable is in group
                ``/forecast/model`` then *groups* would be
                ``['forecast', 'model']``.

        :Returns:

            `netCDF4.Dataset` or `netCDF4.Group`
             or `h5netcdf.File` or `h5netcdf.Group`
                The group object, which might be the root group.

        """
        for g in groups:
            dataset = dataset.groups[g]

        return dataset

    def _set_attributes(self, var):
        """Set the netCDF variable attributes.

        These are set from the netCDF variable attributes, but only if
        they have not already been defined, either during {{class}}
        instantiation or by a previous call to `_set_attributes`.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            var: `netCDF4.Variable` or `h5netcdf.Variable`
                The netCDF variable.

        :Returns:

            `dict`
                The attributes.

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}._set_attributes"
        )  # pragma: no cover

    @property
    def array(self):
        """Return an independent numpy array containing the data.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `numpy.ndarray`
                An independent numpy array of the data.

        **Examples**

        >>> n = numpy.asanyarray(a)
        >>> isinstance(n, numpy.ndarray)
        True

        """
        return self[...]

    def close(self, dataset):
        """Close the dataset containing the data.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            dataset:
                The dataset to be closed.

        :Returns:

            `None`

        """
        if self._get_component("close"):
            dataset.close()

    def get_format(self):
        """The format of the files.

        .. versionadded:: (cfdm) 1.10.1.0

        .. seealso:: `get_address`, `get_filename`, `get_formats`

        :Returns:

            `str`
                The file format. Always ``'nc'``, signifying netCDF.

        **Examples**

        >>> a.get_format()
        'nc'

        """
        return "nc"

    def get_mask(self):
        """Whether or not to automatically mask the data.

        .. versionadded:: (cfdm) 1.8.2

        **Examples**

        >>> b = a.get_mask()

        """
        return self._get_component("mask")

    def get_missing_values(self, default=ValueError()):
        """The missing value indicators from the netCDF variable.

        Deprecated at version 1.11.2.0. Use `get_attributes` instead.

        .. versionadded:: (cfdm) 1.10.0.3

        :Parameters:

            default: optional
                Return the value of the *default* parameter if no missing
                values have yet been defined.

                {{default Exception}}

        :Returns:

            `dict` or `None`
                The missing value indicators from the netCDF variable,
                keyed by their netCDF attribute names. An empty
                dictionary signifies that no missing values are given
                in the file. `None` signifies that the missing values
                have not been set.

        **Examples**

        >>> a.get_missing_values(None)
        None

        >>> b.get_missing_values({})
        {}

        >>> b.get_missing_values()
        {}

        >>> c.get_missing_values()
        {'missing_value': 1e20, 'valid_range': (-10, 20)}

        >>> d.get_missing_values()
        {'valid_min': -999}

        """
        raise DeprecationError(
            f"{self.__class__.__name__}.get_missing_values was deprecated "
            "at version 1.11.2.0 and is no longer available. "
            f"Use {self.__class__.__name__}.get_attributes instead."
        )

    def get_unpack(self):
        """Whether or not to automatically unpack the data.

        .. versionadded:: (cfdm) 1.11.2.0

        **Examples**

        >>> a.get_unpack()
        True

        """
        return self._get_component("unpack")

    def to_memory(self):
        """Bring data on disk into memory.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `NumpyArray`
                The new array with all of its data in memory.

        """
        return NumpyArray(self[...])



from .container import Container
from .files import Files
from .properties import Properties
from .propertiesdata import PropertiesData
from .propertiesdatabounds import PropertiesDataBounds
from .coordinate import Coordinate
from .topology import Topology

from .parameters import Parameters
from .parametersdomainancillaries import ParametersDomainAncillaries

from .netcdf import (
    NetCDFComponents,
    NetCDFGlobalAttributes,
    NetCDFGroupAttributes,
    NetCDFUnlimitedDimension,
    NetCDFDimension,
    NetCDFExternal,
    NetCDFGeometry,
    NetCDFGroupsMixin,
    NetCDFHDF5,
    NetCDFInterpolationSubareaDimension,
    NetCDFMixin,
    NetCDFNodeCoordinateVariable,
    NetCDFSampleDimension,
    NetCDFSubsampledDimension,
    NetCDFUnreferenced,
    NetCDFVariable,
)

from .fielddomain import FieldDomain



import numpy as np

from ....functions import integer_dtype


class PointTopology:
    """Mixin class for point topology array compressed by UGRID.

    Subclasses must also inherit from `MeshSubarray`.

    .. versionadded:: (cfdm) 1.11.0.0

    """

    def __getitem__(self, indices):
        """Return a subspace of the uncompressed data.

        x.__getitem__(indices) <==> x[indices]

        .. versionadded:: (cfdm) 1.11.0.0

        """
        from math import isnan

        from scipy.sparse import csr_array

        start_index = self.start_index
        node_connectivity = self._select_data(check_mask=False)

        masked = np.ma.isMA(node_connectivity)

        largest_node_id = node_connectivity.max()
        if not start_index:
            # Add 1 to remove all zeros (0 is the fill value in the
            # sparse array), first making sure that the datatpe can
            # handle it.
            if largest_node_id == np.iinfo(node_connectivity.dtype).max:
                node_connectivity = node_connectivity.astype(int, copy=False)

            node_connectivity = node_connectivity + 1
            largest_node_id = largest_node_id + 1

        p = 0
        pointers = [0]
        cols = []
        u = []

        pointers_append = pointers.append
        cols_extend = cols.extend
        u_extend = u.extend

        # WARNING (TODO): This loop is a potential performance bottleneck.
        for node in np.unique(node_connectivity).tolist():
            # Find the collection of all nodes that are joined to this
            # node via links in the mesh, including this node itself
            # (which will be at the start of the list).
            nodes = self._connected_nodes(node, node_connectivity, masked)

            n_nodes = len(nodes)
            p += n_nodes
            pointers_append(p)
            cols_extend(range(n_nodes))
            u_extend(nodes)

        u = np.array(u, dtype=integer_dtype(largest_node_id))
        u = csr_array((u, cols, pointers))
        u = u.toarray()

        if any(map(isnan, self.shape)):
            # Store the shape, now that is it known.
            self._set_component("shape", u.shape, copy=False)

        if indices is not Ellipsis:
            u = u[indices]

        # Mask all zeros
        u = np.ma.where(u == 0, np.ma.masked, u)

        if not start_index:
            # Subtract 1 to get back to zero-based node identities
            u -= 1

        return u



import copy
import datetime
import faulthandler
import itertools
import unittest
from unittest.mock import patch

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

# Note: it is important we test on the cfdm logging config rather than the
# generic Python module logging (i.e. 'cfdm.logging' not just 'logging').
# Also, mimic the use in the codebase by using a module-specific logger:
log_name = __name__
logger = cfdm.logging.getLogger(log_name)


DEBUG_MSG = "A major clue to solving the evasive bug"
DETAIL_MSG = "In practice this will be very detailed."
INFO_MSG = "This should be short and sweet"
WARNING_MSG = "Best pay attention to this!"


@cfdm.decorators._manage_log_level_via_verbosity
def decorated_logging_func(verbose=None):
    """Dummy function to log messages at various levels by decorator.

    See also dummyClass.decorated_logging_method which does the same but
    as a method rather than a function.

    """
    logger.debug(DEBUG_MSG)
    logger.detail(DETAIL_MSG)
    logger.info(INFO_MSG)
    logger.warning(WARNING_MSG)


class dummyClass:
    """Dummy class acting as container to test methods.

    This is a special class to test appropriate methods as proper
    instance methods, mirroring their context in the codebase.

    """

    def __init__(self):
        """Constructor for dummyClass."""
        self._list = [1]

        self.debug_message = DEBUG_MSG
        self.detail_message = DETAIL_MSG
        self.info_message = INFO_MSG
        self.warning_message = WARNING_MSG

        self.dummy_string = "foo bar baz"

    def copy(self):
        """Creates a deep copy."""
        return copy.deepcopy(self)  # note a shallow copy is not sufficient

    def func(self, inplace):
        """Dummy function to do something trivial to a mutable object.

        The operation is potentially done in-place as specified by an
        in-place flag.

        """
        if inplace:
            d = self
        else:
            d = self.copy()

        d._list.append(2)

        if inplace:
            d = None
        return d

    @cfdm.decorators._inplace_enabled(False)
    def decorated_func(self, inplace):
        """Dummy function that is 'func' except managed by decorator.

        The decorator manages whether or not the operation is applied
        in-place.

        """
        d = cfdm.decorators._inplace_enabled_define_and_cleanup(self)
        d._list.append(2)
        return d

    def print_or_return_string(self, display=True):
        """Dummy function to either print or return a given string.

        It prints the string if the display argument is True, else it
        returns it.

        """
        string = self.dummy_string

        if display:
            print(string)
        else:
            return string

    @cfdm.decorators._display_or_return
    def print_or_return_string_by_decorator(self, display=True):
        """Equivalent to 'print_or_return_string' but via decorator.

        The decorator manages whether or not to print rather than return
        depending on whether or not the display argument is True.

        """
        return self.dummy_string

    @cfdm.decorators._manage_log_level_via_verbosity
    def decorated_logging_method(self, verbose=None):
        """Method to log messages at various levels by decorator.

        See also decorated_logging_func which does the same but as a
        function rather than a method.

        """
        logger.debug(self.debug_message)
        logger.detail(self.detail_message)
        logger.info(self.info_message)
        logger.warning(self.warning_message)


class decoratorsTest(unittest.TestCase):
    """Test the decorators module."""

    def test_inplace_enabled(self):
        """Test the `_inplace_enabled` decorator."""
        # Note we must initiate separate classes as a list is mutable:
        test_class_1 = dummyClass()
        test_class_2 = dummyClass()

        # Test when not in-place
        res_1 = test_class_1.func(inplace=False)
        res_2 = test_class_2.decorated_func(inplace=False)
        self.assertEqual(test_class_1._list, test_class_2._list)
        self.assertEqual(test_class_2._list, [1])  # as original list untouched
        self.assertEqual(res_1._list, res_2._list)
        self.assertEqual(res_2._list, [1, 2])  # as return d copy, not original

        # Test when in-place
        res_3 = test_class_1.func(inplace=True)
        res_4 = test_class_2.decorated_func(inplace=True)

        self.assertEqual(test_class_1._list, test_class_2._list)
        # As do the operation in-place on the original (class) list object:
        self.assertEqual(test_class_2._list, [1, 2])
        self.assertEqual(res_3, res_4)
        self.assertEqual(res_4, None)  # as return None if inplace=True

    def test_manage_log_level_via_verbosity(self):
        """Test the `_manage_log_level_via_verbosity` decorator."""
        # Order of decreasing severity/verbosity is crucial to one test below
        levels = ["WARNING", "INFO", "DETAIL", "DEBUG"]

        # Note we test assertions on the root logger object, which is the
        # one output overall at runtime, but the specific module logger name
        # should be registered within the log message:
        log_message = [
            f"WARNING:{log_name}:{WARNING_MSG}",
            f"INFO:{log_name}:{INFO_MSG}",
            f"DETAIL:{log_name}:{DETAIL_MSG}",
            f"DEBUG:{log_name}:{DEBUG_MSG}",
        ]

        test_class = dummyClass()
        # 1. First test it works for methods using test_class to test with
        # 2. Then test it works for functions (not bound to a class)
        functions_to_call_to_test = [
            test_class.decorated_logging_method,  # Case 1 as described above
            decorated_logging_func,  # Case 2
        ]

        for level, function_to_call_to_test in itertools.product(
            levels, functions_to_call_to_test
        ):
            cfdm.log_level(level)  # reset to level

            # Default verbose(=None) cases: log_level should determine output
            with self.assertLogs(level=cfdm.log_level().value) as catch:
                function_to_call_to_test()

                for msg in log_message:
                    # log_level should prevent messages less severe appearing:
                    if levels.index(level) >= log_message.index(msg):
                        self.assertIn(msg, catch.output)
                    else:  # less severe, should be effectively filtered out
                        self.assertNotIn(msg, catch.output)

            # Cases where verbose is set; value should override log_level...

            # Highest verbosity case (note -1 == 'DEBUG', highest verbosity):
            # all messages should appear, regardless of global log_level:
            for argument in (-1, "DEBUG", "debug", "Debug", "DeBuG"):
                with self.assertLogs(level=-1) as catch:
                    function_to_call_to_test(verbose=argument)
                    for msg in log_message:
                        self.assertIn(msg, catch.output)

            # Lowest verbosity case ('WARNING' / 1) excluding special case of
            # 'DISABLE' (see note above): only warning messages should appear,
            # regardless of global log_level value set:
            for argument in (1, "WARNING", "warning", "Warning", "WaRning"):
                with self.assertLogs(level=1) as catch:
                    function_to_call_to_test(verbose=argument)
                    for msg in log_message:
                        if msg.split(":")[0] == "WARNING":
                            self.assertIn(msg, catch.output)
                        else:
                            self.assertNotIn(msg, catch.output)

            # Boolean cases for testing backwards compatibility...

            # ... verbose=True should be equivalent to verbose=3 now:
            with self.assertLogs(level=3) as catch:
                function_to_call_to_test(verbose=True)
                for msg in log_message:
                    if msg.split(":")[0] == "DEBUG":
                        self.assertNotIn(msg, catch.output)
                    else:
                        self.assertIn(msg, catch.output)

            # ... verbose=False should be equivalent to verbose=0 now, so
            # test along with 'DISABLE' special case below...

            # Special 'DISABLE' (0) case: note this needs to be last as we
            # reset the log_level to it but need to use 'NOTSET' for the
            # assertLogs level, which sends all log messages through:
            for argument in (0, "DISABLE", "disable", "Disable", "DisAblE"):
                with self.assertLogs(level="NOTSET") as catch:
                    # Note: get 'AssertionError' if don't log anything at all,
                    # so to avoid this and allow check for disabled logging,
                    # first log something then disable and check that no other
                    # messages emerge:
                    logger.info(
                        "Purely to keep 'assertLog' happy: see comment!"
                    )
                    cfdm.log_level("DISABLE")
                    function_to_call_to_test(verbose=argument)
                    for msg in log_message:  # nothing else should be logged
                        self.assertNotIn(msg, catch.output)

            # verbose=False should be equivalent in behaviour to verbose=0
            with self.assertLogs(level="NOTSET") as catch:
                logger.info("Purely to keep 'assertLog' happy: see previous!")
                function_to_call_to_test(verbose=False)
                for msg in log_message:  # nothing else should be logged
                    self.assertNotIn(msg, catch.output)

    @patch("builtins.print")
    def test_display_or_return(self, mock_print):
        """Test the `_display_or_return` decorator."""
        test_class = dummyClass()

        # Compare results with display=False:
        res_1 = test_class.print_or_return_string(display=False)
        res_2 = test_class.print_or_return_string_by_decorator(display=False)
        self.assertEqual(res_1, res_2)
        mock_print.assert_not_called()  # checks nothing was printed to STDOUT

        # Compare results with display=True:
        res_3 = test_class.print_or_return_string_by_decorator(display=True)
        mock_print.assert_called_with(test_class.dummy_string)
        # Should print rather than return, so returns None by default:
        self.assertEqual(res_3, None)

        new_string = "Let's change it up"
        test_class.dummy_string = new_string
        # Compare defaults, where default should be display=True i.e. to print
        res_4 = test_class.print_or_return_string_by_decorator()
        mock_print.assert_called_with(new_string)
        self.assertEqual(res_4, None)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



"""`cfdm.core`, the core reference implementation of the CF data model.

It is a stand-alone core implementation that includes no functionality
beyond that mandated by the CF data model (and therefore excludes any
information about the netCDF encoding of constructs).

The core implementation provides the basis (via inheritance) for the
`cfdm` package that provides further practical functionality in
addition, for example enabling the reading and writing of netCDF
datasets and the inspection of CF data model constructs.

"""

__date__ = "2025-04-01"
__cf_version__ = "1.12"
__version__ = "1.12.1.0"

from platform import python_version

_requires = ("numpy", "packaging")
_error0 = f"cfdm.core requires the modules {', '.join(_requires)}. "

# Check the version of packaging
try:
    import packaging
    from packaging.version import Version
except ImportError as error1:
    raise ImportError(_error0 + str(error1))
else:
    _minimum_vn = "20.0"
    if Version(packaging.__version__) < Version(_minimum_vn):
        raise RuntimeError(
            f"Bad packaging version: cf requires packaging>={_minimum_vn}. "
            f"Got {packaging.__version__} at {packaging.__file__}"
        )

# Check the version of python
_minimum_vn = "3.9.0"
if Version(python_version()) < Version(_minimum_vn):
    raise ValueError(
        f"Bad python version: cfdm.core requires python>={_minimum_vn}. "
        f"Got {python_version()}"
    )

# Check the version of numpy
try:
    import numpy as np
except ImportError as error1:
    raise ImportError(_error0 + str(error1))
else:
    _minimum_vn = "2.0.0"
    if Version(np.__version__) < Version(_minimum_vn):
        raise ValueError(
            f"Bad numpy version: cfdm.core requires numpy>={_minimum_vn}. "
            f"Got {np.__version__} at {np.__file__}"
        )

del _minimum_vn

from .constructs import Constructs

from .functions import CF, environment

from .data import Data, Array, NumpyArray

from .bounds import Bounds
from .coordinateconversion import CoordinateConversion
from .datum import Datum
from .domain import Domain
from .interiorring import InteriorRing

from .auxiliarycoordinate import AuxiliaryCoordinate
from .cellconnectivity import CellConnectivity
from .cellmeasure import CellMeasure
from .cellmethod import CellMethod
from .coordinatereference import CoordinateReference
from .dimensioncoordinate import DimensionCoordinate
from .domainancillary import DomainAncillary
from .domainaxis import DomainAxis
from .domaintopology import DomainTopology
from .field import Field
from .fieldancillary import FieldAncillary

from .abstract import (
    Container,
    Properties,
    PropertiesData,
    PropertiesDataBounds,
    Coordinate,
    Parameters,
    ParametersDomainAncillaries,
    Topology,
)

from .meta import DocstringRewriteMeta



class QuadraticInterpolation:
    """Mixin class for quadratic interpolation.

    See CF appendix J "Coordinate Interpolation Methods".

    .. versionadded:: (cfdm) 1.10.0.0

    """

    def _fq(self, *args, **kwargs):
        """Alias for `_quadratic_interpolation`.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return self._quadratic_interpolation(*args, **kwargs)

    def _fw(self, ua, ub, u_i, d, s_i):
        """Calculate the quadratic interpolation parameter ``w``.

        Calculates the quadratic interpolation parameter ``w`` for a
        subsampled dimension.

        w = fw(ua, ub, u(i), s(i))
          = (u(i) - (1-s(i))*ua - s(i)*ub) / (4*(1-s(i))*s(i))

        See CF appendix J "Coordinate Interpolation Methods".

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `_quadratic_interpolation`

        :Parameters:

            ua, ub: `numpy.ndarray`
                The tie points at locations A and B, in the sense of
                CF appendix J Figure J.1.

            u_i: `numpy.ndarray`
                The uncompressed value at the interior location
                implied by *s_i*.

            {{d: `int`}}

            {{s_i: array_like}}

        :Returns:

            `numpy.ndarray`

        """
        s = self._s(d, s=s_i)
        one_minus_s = 1 - s

        return (u_i - one_minus_s * ua - s * ub) / (4 * one_minus_s * s)

    def _quadratic_interpolation(self, ua, ub, w, d1, s=None):
        """Interpolate quadratically between two of tie points.

        Computes the quadratic interpolation operator ``fq``, where
        ``fl`` is the linear interpolation operator and ``w`` is the
        quadratic coefficient:

        u = fq(ua, ub, w, s)
          = ua + s*(ub - ua + 4*w*(1-s))
          = ua*(1-s) + ub*s + 4*w*s*(1-s)
          = fl(ua, ub, s) + 4*w*s*(1-s)

        See CF appendix J "Coordinate Interpolation Methods".

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `_linear_interpolation`

        :Parameters:

            ua, ub: `numpy.ndarray`
                The tie points at locations A and B, in the sense of
                CF appendix J Figure J.1.

            w: `numpy.ndarray` or `None`
                The quadratic interpolation coefficient, with the same
                number of dimensions in the same relative order as the
                full tie points array. If `None` then assumed to be
                zero.

            {{d1: `int`}}

            {{s: array_like, optional}}

        :Returns:

            `numpy.ndarray`
                The result of interpolating the tie points to interior
                locations implied by *s*.

        """
        s = self._s(d1, s=s)

        if w is not None:
            u = ua + s * (ub - ua + 4 * w * (1 - s))
        else:
            u = ua + s * (ub - ua)

        return u



import logging

from ..decorators import _manage_log_level_via_verbosity
from . import Parameters

logger = logging.getLogger(__name__)


class ParametersDomainAncillaries(Parameters):
    """Mixin class for parameter- and ancillary-valued terms.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __bool__(self):
        """Called by the `bool` built-in function.

        x.__bool__() <==> bool(x)

        .. versionadded:: (cfdm) 1.7.0

        """
        return super().__bool__() or bool(self.domain_ancillaries())

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        """
        out = [super().__str__()]

        x = ", ".join(sorted(self.domain_ancillaries()))
        out.append(f"Ancillaries: {x}")

        return "; ".join(out)

    @_manage_log_level_via_verbosity
    def equals(
        self,
        other,
        rtol=None,
        atol=None,
        verbose=None,
        ignore_data_type=False,
        ignore_fill_value=False,
        ignore_type=False,
    ):
        """Whether two instances are the same.

        Equality is strict by default. This means that:

        * the named parameters must be the same, with the same values
          and data types, and vector-valued parameters must also have
          same the size and be element-wise equal (see the
          *ignore_data_type* parameter), and

        ..

        * the names of domain ancillary constructs are the same.

        {{equals tolerance}}

        Any type of object may be tested but, in general, equality is
        only possible with another object of the same type, or a
        subclass of one. See the *ignore_type* parameter.

        :Parameters:

            other:
                The object to compare for equality.

            {{atol: number, optional}}

            {{rtol: number, optional}}

            {{ignore_type: `bool`, optional}}

            {{verbose: `int` or `str` or `None`, optional}}

        :Returns:

            `bool`
                Whether the two instances are equal.

        **Examples**

        >>> c.equals(c)
        True
        >>> c.equals(c.copy())
        True
        >>> c.equals(None)
        False

        """
        if not super().equals(
            other,
            rtol=rtol,
            atol=atol,
            verbose=verbose,
            ignore_type=ignore_type,
        ):
            return False

        domain_ancillaries0 = self.domain_ancillaries()
        domain_ancillaries1 = other.domain_ancillaries()
        if set(domain_ancillaries0) != set(domain_ancillaries1):
            logger.info(
                f"{self.__class__.__name__}: Different domain ancillary "
                "terms "
                f"({set(domain_ancillaries0)} != {set(domain_ancillaries1)})"
            )  # pragma: no cover
            return False

        for term, value0 in domain_ancillaries0.items():
            value1 = domain_ancillaries1[term]
            if value0 is None and value1 is None:
                continue

            if value0 is None or value1 is None:
                logger.info(
                    f"{self.__class__.__name__}: Unequal {term!r} domain "
                    f"ancillary terms ({value0!r} != {value1!r})"
                )  # pragma: no cover
                return False

        # Still here? Then the two instances are as equal as can be
        # ascertained in the absence of domains.
        return True



from .abstract import SubsampledSubarray
from .mixin import QuadraticLatitudeLongitudeInterpolation


class QuadraticLatitudeLongitudeSubarray(
    QuadraticLatitudeLongitudeInterpolation, SubsampledSubarray
):
    """A subarray of an array compressed by subsamplng.

    A subarray describes a unique part of the uncompressed array.

    The compressed data is reconstituted by quadratic latitude
    longitude interpolation.

    See CF appendix J "Coordinate Interpolation Methods".

    .. versionadded:: (cfdm) 1.10.0.0

    """

    def __getitem__(self, indices):
        """Return a subspace of the uncompressed data.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the uncompressed data as an independent
        numpy array.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        (d1,) = tuple(self.compressed_dimensions())

        lat, lon = self._codependent_tie_points("latitude", "longitude")

        lat = self._select_data(lat)
        lon = self._select_data(lon)

        u = self._quadratic_latitude_longitude_interpolation(
            lat_a=self._select_location(lat, {d1: 0}),
            lon_a=self._select_location(lon, {d1: 0}),
            lat_b=self._select_location(lat, {d1: 1}),
            lon_b=self._select_location(lon, {d1: 1}),
            ce=self._select_parameter("ce"),
            ca=self._select_parameter("ca"),
            location_use_3d_cartesian=self._select_parameter(
                "location_use_3d_cartesian"
            ),
            d1=d1,
        )
        u = self._post_process(u)

        if indices is Ellipsis:
            return u

        return u[indices]



from ..netcdf4array import NetCDF4Array
from .mixin import FragmentFileArrayMixin


class FragmentNetCDF4Array(FragmentFileArrayMixin, NetCDF4Array):
    """A fragment of aggregated data in a file accessed with `netCDF4`.

    .. versionadded:: (cfdm) 1.12.0.0

    """



import datetime
import faulthandler
import os
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

verbose = False
warnings = False


class create_fieldTest_2(unittest.TestCase):
    """Test ab initio creation of field constructs in memory."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.LOG_LEVEL("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')
        self.filename = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test_file_b.nc"
        )

    def test_create_field_2(self):
        """Test ab initio creation of a second variation of field."""
        # Dimension coordinates
        dim1 = cfdm.DimensionCoordinate(data=cfdm.Data(np.arange(10.0)))
        dim1.set_property("standard_name", "projection_y_coordinate")
        dim1.set_property("units", "m")

        data = np.arange(9.0) + 20
        data[-1] = 34
        dim0 = cfdm.DimensionCoordinate(data=cfdm.Data(data))
        dim0.set_property("standard_name", "projection_x_coordinate")
        dim0.set_property("units", "m")

        array = dim0.data.array

        array = np.array([array - 0.5, array + 0.5]).transpose((1, 0))
        array[-2, 1] = 30
        array[-1, :] = [30, 36]
        dim0.set_bounds(cfdm.Bounds(data=cfdm.Data(array)))

        dim2 = cfdm.DimensionCoordinate(
            data=cfdm.Data([1.5]),
            bounds=cfdm.Bounds(data=cfdm.Data([[1, 2.0]])),
        )
        dim2.set_property(
            "standard_name", "atmosphere_hybrid_height_coordinate"
        )

        # Auxiliary coordinates
        aux2 = cfdm.AuxiliaryCoordinate(
            data=cfdm.Data(np.arange(-45, 45, dtype="int32").reshape(10, 9))
        )
        aux2.set_property("units", "degree_N")
        aux2.set_property("standard_name", "latitude")

        aux3 = cfdm.AuxiliaryCoordinate(
            data=cfdm.Data(np.arange(60, 150, dtype="int32").reshape(9, 10))
        )
        aux3.set_property("standard_name", "longitude")
        aux3.set_property("units", "degreeE")

        array = np.ma.array(
            [
                "alpha",
                "beta",
                "gamma",
                "delta",
                "epsilon",
                "zeta",
                "eta",
                "theta",
                "iota",
                "kappa",
            ],
        )
        array[0] = np.ma.masked
        aux4 = cfdm.AuxiliaryCoordinate(data=cfdm.Data(array))
        aux4.set_property("standard_name", "greek_letters")

        # Domain ancillaries
        ak = cfdm.DomainAncillary(data=cfdm.Data([10.0]))
        ak.set_property("units", "m")
        ak.set_bounds(cfdm.Bounds(data=cfdm.Data([[5, 15.0]])))

        bk = cfdm.DomainAncillary(data=cfdm.Data([20.0]))
        bk.set_bounds(cfdm.Bounds(data=cfdm.Data([[14, 26.0]])))

        # Cell measures
        msr0 = cfdm.CellMeasure(
            data=cfdm.Data(1 + np.arange(90.0).reshape(9, 10) * 1234)
        )
        msr0.set_measure("area")
        msr0.set_property("units", "km2")

        # Data
        data = cfdm.Data(np.arange(90.0).reshape(10, 9))

        properties = {"units": "m s-1"}

        f = cfdm.Field(properties=properties)
        f.set_property("standard_name", "eastward_wind")

        axisX = f.set_construct(cfdm.DomainAxis(9))
        axisY = f.set_construct(cfdm.DomainAxis(10))
        axisZ = f.set_construct(cfdm.DomainAxis(1))

        f.set_data(data, axes=[axisY, axisX])

        x = f.set_construct(dim0, axes=[axisX])
        y = f.set_construct(dim1, axes=[axisY])
        z = f.set_construct(dim2, axes=[axisZ])

        lat = f.set_construct(aux2, axes=[axisY, axisX])
        lon = f.set_construct(aux3, axes=[axisX, axisY])
        f.set_construct(aux4, axes=[axisY])

        ak = f.set_construct(ak, axes=[axisZ])
        bk = f.set_construct(bk, axes=[axisZ])

        # Coordinate references
        coordinate_conversion = cfdm.CoordinateConversion(
            parameters={
                "grid_mapping_name": "transverse_mercator",
                "latitude_of_projection_origin": 49.0,
                "longitude_of_central_meridian": -2.0,
                "scale_factor_at_central_meridian": 0.9996012717,
                "false_easting": 400000.0,
                "false_northing": -100000.0,
                "unit": "metre",
            }
        )

        datum0 = cfdm.Datum(
            parameters={
                "inverse_flattening": 299.3249646,
                "longitude_of_prime_meridian": 0.0,
                "semi_major_axis": 6377563.396,
            }
        )

        ref0 = cfdm.CoordinateReference(
            coordinates=[x, y],
            datum=datum0,
            coordinate_conversion=coordinate_conversion,
        )

        coordinate_conversion = cfdm.CoordinateConversion(
            parameters={"grid_mapping_name": "latitude_longitude"}
        )

        datum2 = cfdm.Datum(
            parameters={
                "longitude_of_prime_meridian": 0.0,
                "semi_major_axis": 6378137.0,
                "inverse_flattening": 298.257223563,
            }
        )

        ref2 = cfdm.CoordinateReference(
            coordinates=[lat, lon],
            datum=datum2,
            coordinate_conversion=coordinate_conversion,
        )

        f.set_construct(msr0, axes=[axisX, axisY])

        f.set_construct(ref0)
        f.set_construct(ref2)

        orog = cfdm.DomainAncillary(data=f.get_data())
        orog.set_property("standard_name", "surface_altitude")
        orog.set_property("units", "m")
        orog = f.set_construct(orog, axes=[axisY, axisX])

        coordinate_conversion = cfdm.CoordinateConversion(
            parameters={
                "standard_name": "atmosphere_hybrid_height_coordinate"
            },
            domain_ancillaries={"orog": orog, "a": ak, "b": bk},
        )

        ref1 = cfdm.CoordinateReference(
            coordinates=[z],
            datum=datum0,
            coordinate_conversion=coordinate_conversion,
        )

        f.set_construct(ref1)

        # Field ancillary variables
        g = f.copy()
        anc = cfdm.FieldAncillary(data=g.get_data())
        anc.standard_name = "ancillaryA"
        f.set_construct(anc, axes=[axisY, axisX])

        g = f[0]
        g = g.squeeze()
        anc = cfdm.FieldAncillary(data=g.get_data())
        anc.standard_name = "ancillaryB"
        f.set_construct(anc, axes=[axisX])

        g = f[..., 0]
        g = g.squeeze()
        anc = cfdm.FieldAncillary(data=g.get_data())
        anc.standard_name = "ancillaryC"
        f.set_construct(anc, axes=[axisY])

        f.set_property("flag_values", np.array([1, 2, 4], "int32"))
        f.set_property("flag_meanings", "a bb ccc")
        f.set_property("flag_masks", [2, 1, 0])

        cm0 = cfdm.CellMethod(
            axes=[axisX],
            method="mean",
            qualifiers={"interval": [cfdm.Data(1, "day")], "comment": "ok"},
        )

        cm1 = cfdm.CellMethod(
            axes=[axisY], method="maximum", qualifiers={"where": "sea"}
        )

        f.set_construct(cm0)
        f.set_construct(cm1)

        self.assertTrue(
            f.equals(f.copy(), verbose=verbose),
            "Field f not equal to a copy of itself",
        )

        for fmt in (
            "NETCDF3_CLASSIC",
            "NETCDF3_64BIT",
            "NETCDF4",
            "NETCDF4_CLASSIC",
        ):
            cfdm.write(f, self.filename, fmt=fmt, verbose=verbose)

            g = cfdm.read(self.filename, verbose=verbose)

            self.assertEqual(len(g), 1, f"{len(g)} != 1")

            g = g[0].squeeze()

            self.assertEqual(
                sorted(f.constructs),
                sorted(g.constructs),
                f"\n\nf (created in memory)"
                f"\n{f.constructs}"
                f"\n\n{f.constructs.items()}"
                f"\n\ng (read from disk)"
                f"\n{g.constructs}"
                f"\n\n{g.constructs.items()}",
            )

            self.assertTrue(
                g.equals(g.copy(), verbose=verbose),
                "Field g not equal to a copy of itself",
            )

            self.assertTrue(
                g.equals(f, verbose=verbose),
                "Field not equal to itself read back in",
            )

        x = g.dump(display=False)
        x = f.dump(display=False)

        g = cfdm.read(
            self.filename,
            verbose=verbose,
            extra=["domain_ancillary"],
            warnings=warnings,
        )


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



import numpy as np

from .abstract import MeshSubarray
from .mixin import PointTopology


class PointTopologyFromEdgesSubarray(PointTopology, MeshSubarray):
    """A subarray of a point topology array compressed by UGRID edges.

    A subarray describes a unique part of the uncompressed array.

    .. versionadded:: (cfdm) 1.11.0.0

    """

    def _connected_nodes(self, node, node_connectivity, masked):
        """Return the nodes that are joined to *node* by edges.

        The input *node* is included at the start of the returned
        list.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            node: `int`
                A node identifier.

            node_connectivity: `numpy.ndarray`
                A UGRID "edge_node_connectivity" array.

            masked: `bool`, optional
                Whether or not *node_connectivity* has masked
                elements.

        :Returns:

            `list`
                All nodes that are joined to *node*, including *node*
                itself at the start.

        """
        nodes = sorted(
            set(
                node_connectivity[np.where(node_connectivity == node)[0]]
                .flatten()
                .tolist()
            )
        )

        # Move 'node' to the front of the list
        nodes.remove(node)
        nodes.insert(0, node)

        return nodes



from ..data import Data
from .properties import Properties


class PropertiesData(Properties):
    """Mixin class for a data array with descriptive properties.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __new__(cls, *args, **kwargs):
        """Store component classes.

        NOTE: If a child class requires a different component classes
        than the ones defined here, then they must be redefined in the
        child class.

        """
        instance = super().__new__(cls)
        instance._Data = Data
        return instance

    def __init__(
        self,
        properties=None,
        data=None,
        source=None,
        copy=True,
        _use_data=True,
    ):
        """**Initialisation**

        :Parameters:

            {{init properties: `dict`, optional}}

                *Parameter example:*
                  ``properties={'standard_name': 'altitude'}``

            {{init data: data_like, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(properties=properties, source=source, copy=copy)

        if source is not None:
            if not _use_data:
                data = None
            else:
                try:
                    data = source.get_data(None)
                except AttributeError:
                    data = None

        if _use_data and data is not None:
            self.set_data(data, copy=copy)

    def __array__(self, dtype=None, copy=None):
        """The numpy array interface.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            dtype: optional
                Typecode or data-type to which the array is cast.

            copy: `None` or `bool`
                Included to match the v2 `numpy.ndarray.__array__`
                API, but ignored. The returned numpy array is always
                independent.

                .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `numpy.ndarray`
                An independent numpy array of the data.

        """
        data = self.get_data(None)
        if data is not None:
            return data.__array__(dtype, copy=copy)

        raise ValueError(f"{self.__class__.__name__} has no data")

    def __data__(self):
        """Defines the data interface.

        Returns the data, if any.

        :Returns:

            `Data`

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_data([1, 2, 3])
        >>> f.set_property('units', 'K')
        >>> d = {{package}}.Data(f)
        >>> d
        <{{repr}}Data(3): [1, 2, 3] K>

        """
        data = self.get_data(None)
        if data is not None:
            return data

        raise ValueError(f"{self.__class__.__name__} has no data")

    @property
    def data(self):
        """The data.

        ``f.data`` is equivalent to ``f.get_data()``.

        Note that a `Data` instance is returned. Use the `array`
        attribute to get the data as a `numpy` array.

        The units, calendar and fill value properties are, if set,
        inserted into the data.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `{{package}}.Data.array`, `del_data`, `get_data`,
                     `has_data`, `set_data`

        :Returns:

            `Data`
                The data.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_data({{package}}.Data(numpy.arange(10.)))
        >>> f.has_data()
        True
        >>> d = f.data
        >>> d
        <{{repr}}Data(10): [0.0, ..., 9.0]>
        >>> f.data.shape
        (10,)

        """
        return self.get_data()

    @data.setter
    def data(self, value):
        raise AttributeError(
            "Can't set attribute 'data'. Use the 'set_data' method, "
            "or assignment by indexing."
        )

    @data.deleter
    def data(self):
        self.del_data()

    @property
    def dtype(self):
        """Data-type of the data elements.

        **Examples**

        >>> d.dtype
        dtype('float64')
        >>> type(d.dtype)
        <type 'numpy.dtype'>

        """
        data = self.get_data(None, _units=False, _fill_value=False)
        if data is not None:
            return data.dtype

        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute 'dtype'"
        )

    @property
    def ndim(self):
        """The number of data dimensions.

        Only dimensions that correspond to domain axis constructs are
        included.

        .. seealso:: `data`, `has_data`, `isscalar`, `shape`, `size`

        **Examples**

        >>> f.shape
        (73, 96)
        >>> f.ndim
        2
        >>> f.size
        7008

        >>> f.shape
        (73, 1, 96)
        >>> f.ndim
        3
        >>> f.size
        7008

        >>> f.shape
        (73,)
        >>> f.ndim
        1
        >>> f.size
        73

        >>> f.shape
        ()
        >>> f.ndim
        0
        >>> f.size
        1

        """
        data = self.get_data(None, _units=False, _fill_value=False)
        if data is not None:
            return data.ndim

        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute 'ndim'"
        )

    @property
    def shape(self):
        """A tuple of the data array's dimension sizes.

        Only dimensions that correspond to domain axis constructs are
        included.

        .. seealso:: `data`, `has_data`, `ndim`, `size`

        **Examples**

        >>> f.shape
        (73, 96)
        >>> f.ndim
        2
        >>> f.size
        7008

        >>> f.shape
        (73, 1, 96)
        >>> f.ndim
        3
        >>> f.size
        7008

        >>> f.shape
        (73,)
        >>> f.ndim
        1
        >>> f.size
        73

        >>> f.shape
        ()
        >>> f.ndim
        0
        >>> f.size
        1

        """
        data = self.get_data(None, _units=False, _fill_value=False)
        if data is not None:
            return data.shape

        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute 'shape'"
        )

    @property
    def size(self):
        """The number elements in the data.

        `size` is equal to the product of `shape`, that only includes
        the sizes of dimensions that correspond to domain axis
        constructs.

        .. seealso:: `data`, `has_data`, `ndim`, `shape`

        **Examples**

        >>> f.shape
        (73, 96)
        >>> f.ndim
        2
        >>> f.size
        7008

        >>> f.shape
        (73, 1, 96)
        >>> f.ndim
        3
        >>> f.size
        7008

        >>> f.shape
        (73,)
        >>> f.ndim
        1
        >>> f.size
        73

        >>> f.shape
        ()
        >>> f.ndim
        0
        >>> f.size
        1

        """
        data = self.get_data(None, _units=False, _fill_value=False)
        if data is not None:
            return data.size

        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute 'size'"
        )

    def copy(self, data=True):
        """Return a deep copy.

        ``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

        Arrays within `Data` instances are copied with a
        copy-on-write technique. This means that a copy takes up very
        little extra memory, even when the original contains very large
        data arrays, and the copy operation is fast.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            data: `bool`, optional
                If True (the default) then copy data, else the
                data is not copied.

        :Returns:

            `{{class}}`
                The deep copy.

        **Examples**

        >>> g = f.copy()
        >>> g = f.copy(data=False)
        >>> g.has_data()
        False

        """
        return type(self)(source=self, copy=True, _use_data=data)

    def del_data(self, default=ValueError()):
        """Remove the data.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `data`, `get_data`, `has_data`, `set_data`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if data have
                not been set.

                {{default Exception}}

        :Returns:

            `Data`
                The removed data.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_data([1, 2, 3])
        >>> f.has_data()
        True
        >>> f.get_data()
        <{{repr}}Data(3): [1, 2, 3]>
        >>> f.data
        <{{repr}}Data(3): [1, 2, 3]>
        >>> f.del_data()
        <{{repr}}Data(3): [1, 2, 3]>
        >>> g = f.set_data([4, 5, 6], inplace=False)
        >>> g.data
        <{{repr}}Data(3): [4, 5, 6]>
        >>> f.has_data()
        False
        >>> print(f.get_data(None))
        None
        >>> print(f.del_data(None))
        None

        """
        return self._del_component("data", default=default)

    def get_data(self, default=ValueError(), _units=True, _fill_value=True):
        """Return the data.

        Note that a `Data` instance is returned. Use its `array` attribute
        to return the data as an independent `numpy` array.

        The units, calendar and fill value properties are, if set,
        inserted into the data.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `{{package}}.Data.array`, `data`, `del_data`,
                     `has_data`, `set_data`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if data have
                not been set.

                {{default Exception}}

        :Returns:

            `Data`
                The data.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_data([1, 2, 3])
        >>> f.has_data()
        True
        >>> f.get_data()
        <{{repr}}Data(3): [1, 2, 3]>
        >>> f.data
        <{{repr}}Data(3): [1, 2, 3]>
        >>> f.del_data()
        <{{repr}}Data(3): [1, 2, 3]>
        >>> g = f.set_data([4, 5, 6], inplace=False)
        >>> g.data
        <{{repr}}Data(3): [4, 5, 6]>
        >>> f.has_data()
        False
        >>> print(f.get_data(None))
        None
        >>> print(f.del_data(None))
        None

        """
        data = self._get_component("data", None)
        if data is None:
            if default is None:
                return

            return self._default(
                default, message=f"{self.__class__.__name__} has no data"
            )

        if _units:
            # Copy the parent units and calendar to the data
            units = self.get_property("units", None)
            if units is not None:
                data.set_units(units)
            else:
                data.del_units(default=None)

            calendar = self.get_property("calendar", None)
            if calendar is not None:
                data.set_calendar(calendar)
            else:
                data.del_calendar(default=None)

        if _fill_value:
            # Copy the fill_value to the data
            fill_value = self.get_property(
                "missing_value", self.get_property("_FillValue", None)
            )
            if fill_value is not None:
                data.set_fill_value(fill_value)
            else:
                data.del_fill_value(default=None)

        return data

    def has_bounds(self):
        """Whether or not there are cell bounds.

        This is always False.

        .. versionadded:: (cfdm) 1.7.4

        .. seealso:: `has_data`

        :Returns:

            `bool`
                Always False.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.has_bounds()
        False

        """
        return False

    def has_data(self):
        """Whether or not the construct has data.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `data`, `del_data`, `get_data`, `set_data`

        :Returns:

            `bool`
                True if data have been set, otherwise False.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_data([1, 2, 3])
        >>> f.has_data()
        True
        >>> f.get_data()
        <{{repr}}Data(3): [1, 2, 3]>
        >>> f.data
        <{{repr}}Data(3): [1, 2, 3]>
        >>> f.del_data()
        <{{repr}}Data(3): [1, 2, 3]>
        >>> g = f.set_data([4, 5, 6], inplace=False)
        >>> g.data
        <{{repr}}Data(3): [4, 5, 6]>
        >>> f.has_data()
        False
        >>> print(f.get_data(None))
        None
        >>> print(f.del_data(None))
        None

        """
        return self._has_component("data")

    def set_data(self, data, copy=True, inplace=True):
        """Set the data.

        The units, calendar and fill value of the incoming `Data` instance
        are removed prior to insertion.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `data`, `del_data`, `get_data`, `has_data`

        :Parameters:

            data: data_like
                The data to be inserted.

                {{data_like}}

            copy: `bool`, optional
                If True (the default) then copy the data prior to
                insertion, else the data is not copied.

            {{inplace: `bool`, optional (default True)}}

                .. versionadded:: (cfdm) 1.8.7.0

        :Returns:

            `None` or `{{class}}`
                If the operation was in-place then `None` is returned,
                otherwise return a new `{{class}}` instance containing the
                new data.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_data([1, 2, 3])
        >>> f.has_data()
        True
        >>> f.get_data()
        <{{repr}}Data(3): [1, 2, 3]>
        >>> f.data
        <{{repr}}Data(3): [1, 2, 3]>
        >>> f.del_data()
        <{{repr}}Data(3): [1, 2, 3]>
        >>> g = f.set_data([4, 5, 6], inplace=False)
        >>> g.data
        <{{repr}}Data(3): [4, 5, 6]>
        >>> f.has_data()
        False
        >>> print(f.get_data(None))
        None
        >>> print(f.del_data(None))
        None

        """
        _Data = self._Data
        if not isinstance(data, _Data):
            data = _Data(data, copy=False)

        if copy:
            data = data.copy()

        if inplace:
            f = self
        else:
            f = self.copy(data=False)

        f._set_component("data", data, copy=False)

        if inplace:
            return

        return f



from itertools import accumulate, product
from numbers import Number

import numpy as np

from .abstract import CompressedArray
from .mixin import CompressedArrayMixin
from .subarray import GatheredSubarray


class DeprecationError(Exception):
    """Deprecation error."""

    pass


class GatheredArray(CompressedArrayMixin, CompressedArray):
    """An underlying gathered array.

    Compression by gathering combines axes of a multidimensional array
    into a new, discrete axis whilst omitting the missing values and
    thus reducing the number of values that need to be stored.

    The information needed to uncompress the data is stored in a "list
    variable" that gives the indices of the required points.

    See CF section 8.2 "Lossless Compression by Gathering".

    .. versionadded:: (cfdm) 1.7.0

    """

    def __new__(cls, *args, **kwargs):
        """Store subarray classes.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        instance = super().__new__(cls)
        instance._Subarray = {"gathered": GatheredSubarray}
        return instance

    def __init__(
        self,
        compressed_array=None,
        shape=None,
        size=None,
        ndim=None,
        compressed_dimension=None,
        compressed_dimensions=None,
        list_variable=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            compressed_array: array_like
                The compressed array.

            shape: `tuple`
                The shape of the uncompressed array.

            {{init compressed_dimensions: `dict`}}

                *Parameter example:*
                  ``{2: (2, 3)}``

                .. versionadded:: (cfdm) 1.10.0.0

            list_variable: `List`
                The "list variable" required to uncompress the data,
                identical to the data of a CF-netCDF list variable.

            compressed_dimension: deprecated at version 1.10.0.0
                Use the *compressed_dimensions* parameter instead.

            {{init source: optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            {{init copy: `bool`, optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            size: `int`
                Deprecated at version 1.10.0.0. Ignored if set.

                Number of elements in the uncompressed array.

            ndim: `int`
                Deprecated at version 1.10.0.0. Ignored if set.

                The number of uncompressed array dimensions

        """
        if compressed_dimension is not None:
            raise DeprecationError(
                "The 'compressed_dimension' keyword was deprecated at "
                "version 1.10.0.0. "
                "Use the 'compressed_dimensions' keyword instead."
            )  # pragma: no cover

        super().__init__(
            compressed_array=compressed_array,
            shape=shape,
            compressed_dimensions=compressed_dimensions,
            compression_type="gathered",
            source=source,
            copy=copy,
        )

        if source is not None:
            try:
                list_variable = source.get_list(None)
            except AttributeError:
                list_variable = None

        if list_variable is not None:
            self._set_component("list_variable", list_variable, copy=copy)

    def _uncompressed_indices(self):
        """Indices of the uncompressed subarray for the compressed data.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `tuple`
                The indices of the uncompressed subarray for the
                compressed data. Dimensions not compressed by
                gathering will have an index of ``slice(None)``.

        **Examples**

        For an original 3-d array with shape (12, 4, 6) for which the
        last two dimensions have been compressed by gathering with
        list variable indices (1, 2, 5, 6, 13, 15, 16, 22) then:

        >>> for i in x._uncompressed_indices():
        ...     print(i)
        ...
        slice(None, None, None)
        array([0, 0, 0, 1, 2, 2, 2, 3])
        array([1, 2, 5, 0, 1, 3, 4, 4])

        """
        _, u_dims = self.compressed_dimensions().popitem()
        list_variable = np.array(self.get_list())

        u_indices = [slice(None)] * self.ndim

        u_indices[u_dims[0] : u_dims[-1] + 1] = np.unravel_index(
            list_variable, self.shape[u_dims[0] : u_dims[-1] + 1]
        )

        return tuple(u_indices)

    @property
    def dtype(self):
        """The data-type of the uncompressed data.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return self.source().dtype

    def conformed_data(self):
        """The compressed data and list variable.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `dict`
                The conformed gathered data, with the key ``'data'``;
                and the `tuple` of uncompressed indices with the key
                ``'uncompressed_indices'``.

        """
        out = super().conformed_data()
        out["uncompressed_indices"] = self._uncompressed_indices()
        return out

    def get_list(self, default=ValueError()):
        """Return the list variable for a compressed array.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the list
                variable has not been set.

                {{default Exception}}

        :Returns:

            `List`
                The list variable.

        """
        return self._get_component("list_variable", default=default)

    def subarray_shapes(self, shapes):
        """Create the subarray shapes along each uncompressed dimension.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `subarray`

        :Parameters:

            {{subarray_shapes chunks: `int`, sequence, `dict`, or `str`, optional}}

        :Returns:

            `list`
                The subarray sizes along each uncompressed dimension.

        **Examples**

        >>> a.shape
        (2, 3, 4)
        >>> a.compressed_dimensions()
        {0: (0, 1)}
        >>> a.subarray_shapes(-1)
        [(2,), (3,), (4,)]
        >>> a.subarray_shapes("auto")
        [(2,), (3,), "auto"]
        >>> a.subarray_shapes(2)
        [(2,), (3,), 2]
        >>> a.subarray_shapes("60B")
        [(2,), (3,), "60B"]
        >>> a.subarray_shapes((None, None, 2))
        [(2,), (3,), 2]
        >>> a.subarray_shapes((None, None, (1, 3)))
        [(2,), (3,), (1, 3)]
        >>> a.subarray_shapes((None, None, "auto"))
        [(2,), (3,), "auto"]
        >>> a.subarray_shapes((None, None, "60B"))
        [(2,), (3,), "60B"]
        >>> a.subarray_shapes({2: (1, 3)})
        [(2,), (3,), (1, 3)]

        >>> import dask.array as da
        >>> da.core.normalize_chunks(
        ...   a.subarray_shapes("auto"), shape=a.shape, dtype=a.dtype
        ... )
        [(2,), (3,), (4,)]
        >>> da.core.normalize_chunks(
        ...   a.subarray_shapes(2, shape=a.shape, dtype=a.dtype
        ... )
        [(2,), (3,), (2, 2)]

        """
        u_dims = self.get_compressed_axes()

        if shapes == -1:
            return [(size,) for size in self.shape]

        if isinstance(shapes, (str, Number)):
            return [
                (size,) if i in u_dims else shapes
                for i, size in enumerate(self.shape)
            ]

        if isinstance(shapes, dict):
            shapes = [
                shapes[i] if i in shapes else None for i in range(self.ndim)
            ]
        elif len(shapes) != self.ndim:
            raise ValueError(
                f"Wrong number of 'shapes' elements in {shapes}: "
                f"Got {len(shapes)}, expected {self.ndim}"
            )

        # chunks is a sequence
        return [
            (size,) if i in u_dims else c
            for i, (size, c) in enumerate(zip(self.shape, shapes))
        ]

    def subarrays(self, shapes=-1):
        """Return descriptors for every subarray.

        These descriptors are used during subarray decompression.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            {{subarrays chunks: ``-1`` or sequence, optional}}

        :Returns:

             4-`tuple` of iterators
                Each iterable iterates over a particular descriptor
                from each subarray.

                1. The indices of the uncompressed array that
                   correspond to each subarray.

                2. The shape of each uncompressed subarray.

                3. The indices of the compressed array that correspond
                   to each subarray.

                4. The location of each subarray on the uncompressed
                   dimensions.

        **Examples**

        An original 3-d array with shape (4, 73, 96) has been
        compressed by gathering the dimensions with sizes 73 and 96
        respectively into a single dimension of size 3028.

        >>> u_indices, u_shapes, c_indices, locations = x.subarrays()
        >>> for i in u_indices:
        ...    print(i)
        ...
        (slice(0, 4, None), slice(0, 73, None), slice(0, 96, None))
        >>> for i in u_shapes
        ...    print(i)
        ...
        (4, 73, 96)
        >>> for i in c_indices:
        ...    print(i)
        ...
        (slice(0, 4, None), slice(0, 3028, None))
        >>> for i in locations:
        ...    print(i)
        ...
        (0, 0, 0)

        >>> (
        ...  u_indices, u_shapes, c_indices, locations
        ... )= x.subarrays(shapes="most")
        >>> for i in u_indices:
        ...    print(i)
        ...
        (slice(0, 1, None), slice(0, 73, None), slice(0, 96, None))
        (slice(1, 2, None), slice(0, 73, None), slice(0, 96, None))
        (slice(2, 3, None), slice(0, 73, None), slice(0, 96, None))
        (slice(3, 4, None), slice(0, 73, None), slice(0, 96, None))
        >>> for i in u_shapes
        ...    print(i)
        ...
        (1, 73, 96)
        (1, 73, 96)
        (1, 73, 96)
        (1, 73, 96)
        >>> for i in c_indices:
        ...    print(i)
        ...
        (slice(0, 1, None), slice(0, 3028, None))
        (slice(1, 2, None), slice(0, 3028, None))
        (slice(2, 3, None), slice(0, 3028, None))
        (slice(3, 4, None), slice(0, 3028, None))
        >>> for i in locations:
        ...    print(i)
        ...
        (0, 0, 0)
        (1, 0, 0)
        (2, 0, 0)
        (3, 0, 0)

        >>> (
        ...  u_indices, u_shapes, c_indices, locations
        ... ) = x.subarrays(shapes=((3, 1), None, None))
        >>> for i in u_indices:
        ...    print(i)
        ...
        (slice(0, 3, None), slice(0, 73, None), slice(0, 96, None))
        (slice(3, 4, None), slice(0, 73, None), slice(0, 96, None))
        >>> for i in u_shapes
        ...    print(i)
        ...
        (3, 73, 96)
        (1, 73, 96)
        >>> for i in c_indices:
        ...    print(i)
        ...
        (slice(0, 3, None), slice(0, 3028, None))
        (slice(3, 4, None), slice(0, 3028, None))
        >>> for i in locations:
        ...    print(i)
        ...
        (0, 0, 0)
        (1, 0, 0)

        """
        d1, u_dims = self.compressed_dimensions().popitem()

        shapes = self.subarray_shapes(shapes)

        # The indices of the uncompressed array that correspond to
        # each subarray, the shape of each uncompressed subarray, and
        # the location of each subarray
        locations = []
        u_shapes = []
        u_indices = []
        for d, (size, c) in enumerate(zip(self.shape, shapes)):
            if d in u_dims:
                locations.append((0,))
                u_shapes.append((size,))
                u_indices.append((slice(0, size),))
            else:
                locations.append([i for i in range(len(c))])
                u_shapes.append(c)

                c = tuple(accumulate((0,) + c))
                u_indices.append([slice(i, j) for i, j in zip(c[:-1], c[1:])])

        # The indices of the compressed array that correspond to each
        # subarray
        c_indices = []
        for d, size in enumerate(self.source().shape):
            if d == d1:
                c_indices.append((slice(0, size),))
            else:
                if d < d1:
                    c = shapes[d]
                else:
                    c = shapes[d + len(u_dims) - 1]

                c = tuple(accumulate((0,) + c))
                c_indices.append([slice(i, j) for i, j in zip(c[:-1], c[1:])])

        return (
            product(*u_indices),
            product(*u_shapes),
            product(*c_indices),
            product(*locations),
        )

    def to_memory(self):
        """Bring data on disk into memory.

        There is no change to data that is already in memory.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `{{class}}`
                A copy of the array with all of its data in memory.

        """
        a = super().to_memory()

        list_variable = self.get_list(None)
        if list_variable is not None:
            try:
                a._set_component(
                    "list_variable", list_variable.to_memory(), copy=False
                )
            except AttributeError:
                pass

        return a



import numpy as np

from ...functions import integer_dtype
from .abstract import MeshSubarray


class CellConnectivitySubarray(MeshSubarray):
    """A cell connectivity subarray defined by UGRID connectivity.

    A subarray describes a unique part of the uncompressed array.

    A UGRID connectivity variable contains indices which map each cell
    to its neighbours, as found in a UGRID "face_face_connectivity"
    variable.

    .. versionadded:: (cfdm) 1.11.0.0

    .. seealso:: `CellConnectivityArray`

    """

    def __getitem__(self, indices):
        """Return a subspace of the uncompressed data.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the uncompressed data as an independent
        numpy array.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        start_index = self.start_index
        shape = self.shape
        start = 0
        stop = shape[0]
        if start_index:
            start += 1
            stop += 1

        data = self._select_data(check_mask=True)
        if np.ma.isMA(data):
            empty = np.ma.empty
        else:
            empty = np.empty

        dtype = integer_dtype(stop - 1)
        u = empty(shape, dtype=dtype)

        # Store the cell identifiers in the first column
        u[:, 0] = np.arange(start, stop, dtype=dtype)
        # Store the identifiers of the connected cells in the other
        # columns
        u[:, 1:] = data

        if indices is not Ellipsis:
            u = u[indices]

        # Make sure that the values are zero-based
        if start_index:
            u -= 1

        return u



import logging

import numpy as np

from ..decorators import _manage_log_level_via_verbosity
from ..docstring import _docstring_substitution_definitions
from ..functions import atol, rtol

logger = logging.getLogger(__name__)


class Container:
    """Mixin class for storing object components.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        .. versionadded:: (cfdm) 1.7.0

        """
        return f"<{self.__class__.__name__}: {self}>"

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        .. versionadded:: (cfdm) 1.7.0

        """
        out = sorted(self._components)
        return ", ".join(out)

    def __docstring_substitutions__(self):
        """Define docstring substitutions for the class hierarchy.

        The defined substitutions apply to this class along with all
        of its subclasses.

        These are in addtion to, and take precendence over, docstring
        substitutions defined by the base classes of this class.

        See `_docstring_substitutions` for details.

        .. versionaddedd:: (cfdm) 1.8.7.0

        :Returns:

            `dict`
                The docstring substitutions that have been applied.

        """
        return _docstring_substitution_definitions

    def __docstring_package_depth__(self):
        """Return the package depth for {{package}} substitutions.

        See `_docstring_package_depth` for details.

        """
        return 0

    @property
    def _atol(self):
        """Internal alias for `{{package}}.atol`.

        An alias is necessary to avoid a name clash with the keyword
        argument of identical name (`atol`) in calling functions.

        """
        return atol().value

    @property
    def _rtol(self):
        """Internal alias for `{{package}}.rtol`.

        An alias is necessary to avoid a name clash with the keyword
        argument of identical name (`rtol`) in calling functions.

        """
        return rtol().value

    def _equals(
        self,
        x,
        y,
        rtol=None,
        atol=None,
        ignore_data_type=False,
        basic=False,
        **kwargs,
    ):
        """Whether two objects are the same.

        Equality either uses one or other of the objects `!equals`
        methods, or casts them as numpy arrays and carried aout
        numerically tolerant equality checks.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            {{atol: number, optional}}

            {{rtol: number, optional}}

        """
        if basic:
            # x and y can be compared with the basic == operator
            return bool(x == y)

        if rtol is None:
            rtol = self._rtol
        else:
            rtol = float(rtol)

        if atol is None:
            atol = self._atol
        else:
            atol = float(atol)

        kwargs["ignore_data_type"] = ignore_data_type
        kwargs["rtol"] = rtol
        kwargs["atol"] = atol

        eq = getattr(x, "equals", None)
        if callable(eq):
            # --------------------------------------------------------
            # x has a callable "equals" method
            # --------------------------------------------------------
            return eq(y, **kwargs)

        eq = getattr(y, "equals", None)
        if callable(eq):
            # --------------------------------------------------------
            # y has a callable "equals" method
            # --------------------------------------------------------
            return eq(x, **kwargs)

        x = np.asanyarray(x)
        y = np.asanyarray(y)
        if x.shape != y.shape:
            return False

        #        # ------------------------------------------------------------
        #        # Cast x and y as numpy arrays
        #        # ------------------------------------------------------------
        #        if not isinstance(x, np.ndarray):
        #            x = np.asanyarray(x)
        #
        #        if not isinstance(y, np.ndarray):
        #            y = np.asanyarray(y)

        # THIS IS WHERE SOME NUMPY FUTURE WARNINGS ARE COMING FROM

        if not ignore_data_type and x.dtype != y.dtype:
            if x.dtype.kind not in ("S", "U") and y.dtype.kind not in (
                "S",
                "U",
            ):
                return False

        x_is_masked = np.ma.isMA(x)
        y_is_masked = np.ma.isMA(y)

        if not (x_is_masked or y_is_masked):
            try:
                return bool(np.allclose(x, y, rtol=rtol, atol=atol))
            except (IndexError, NotImplementedError, TypeError):
                return bool(np.all(x == y))
        else:
            if x_is_masked and y_is_masked:
                if (x.mask != y.mask).any():
                    return False
            elif (x_is_masked and x.mask.any()) or (
                y_is_masked and y.mask.any()
            ):
                return False

            try:
                return bool(np.ma.allclose(x, y, rtol=rtol, atol=atol))
            except (IndexError, NotImplementedError, TypeError):
                out = np.ma.all(x == y)
                if out is np.ma.masked:
                    return True
                else:
                    return bool(out)

    @_manage_log_level_via_verbosity
    def _equals_preprocess(self, other, verbose=None, ignore_type=False):
        """Common preprocessing prior to testing of equality.

        * If the LHS operand is (object identity) the RHS operand then
          return True.

        * If ignore_type=False and the LHS operand is not of exactly
          the same type as the RHS operand then return False.

        * If ignore_type=True and the LHS operand is not of exactly
          the same type as the RHS operand then instantiate a new
          instance based on the the RHS class and return it.

        .. versionadded:: (cfdm) 1.7.0

        """
        # Check for object identity
        if self is other:
            return True

        # Check that each object is of compatible type
        if ignore_type:
            if not isinstance(other, self.__class__):
                other = type(self)(source=other, copy=False)
        elif not isinstance(other, self.__class__):
            logger.info(
                f"{self.__class__.__name__}: Incompatible type: {type(other)}"
            )
            return False

        return other

    def _iter(self, body, pre=None, post=None, short=False, **kwargs):
        """General purpose iterator.

        .. versionadded:: (cfdm) 1.8.9.0

        :Parameters:

            body: iterable
               An iterable to step through.

            pre: sequence of iterable, optional
               Iterables to step through before *body* is iterated.

            post: sequence of iterable, optional
               Iterables to step through after *body* has been
               iterated.

            short: `bool`, optional
                If True then stop after the first element of either
                *pre*, *body* or *post* is realised.

            kwargs: optional
                Ignored.

        :Return:

            generator
                The elements of *pre*, *body* and *post*.

        """
        if pre:
            for it in pre:
                for x in it:
                    yield x
                    if short:
                        return

        for x in body:
            yield x
            if short:
                return

        if post:
            for it in post:
                for x in it:
                    yield x
                    if short:
                        return

    def _package(self):
        """Return the name of the package in which this class resides.

        :Returns:

            `str`
               The package name

        """
        depth = self.__class__._docstring_package_depth(self.__class__)
        return ".".join(self.__module__.split(".")[0 : depth + 1])



from . import abstract


class Constructs(abstract.Container):
    """A container for metadata constructs.

    The following metadata constructs can be included:

    * domain axis constructs
    * dimension coordinate constructs
    * auxiliary coordinate constructs
    * coordinate reference constructs
    * domain ancillary constructs
    * cell measure constructs
    * domain topology constructs
    * cell connectivity constructs
    * cell method constructs
    * field ancillary constructs

    The container is used by used by `Field` and `Domain` instances.

    The container is like a dictionary in many ways, in that it stores
    key/value pairs where the key is the unique construct key with
    corresponding metadata construct value, and provides some of the
    usual dictionary methods.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        auxiliary_coordinate=None,
        dimension_coordinate=None,
        domain_ancillary=None,
        field_ancillary=None,
        cell_measure=None,
        coordinate_reference=None,
        domain_axis=None,
        domain_topology=None,
        cell_connectivity=None,
        cell_method=None,
        source=None,
        copy=True,
        _use_data=True,
        _view=False,
        _ignore=(),
    ):
        """**Initialisation**

        :Parameters:

            auxiliary_coordinate: `str`, optional
                The base name for keys of auxiliary coordinate
                constructs.

                *Parameter example:*
                  ``auxiliary_coordinate='auxiliarycoordinate'``

            dimension_coordinate: `str`, optional
                The base name for keys of dimension coordinate
                constructs.

                *Parameter example:*
                  ``dimension_coordinate='dimensioncoordinate'``

            domain_ancillary: `str`, optional
                The base name for keys of domain ancillary constructs.

                *Parameter example:*
                  ``domain_ancillary='domainancillary'``

            field_ancillary: `str`, optional
                The base name for keys of field ancillary constructs.

                *Parameter example:*
                  ``field_ancillary='fieldancillary'``

            cell_measure: `str`, optional
                The base name for keys of cell measure constructs.

                *Parameter example:*
                  ``cell_measure='cellmeasure'``

            coordinate_reference: `str`, optional
                The base name for keys of coordinate reference
                constructs.

                *Parameter example:*
                  ``coordinate_reference='coordinatereference'``

            domain_axis: `str`, optional
                The base name for keys of domain axis constructs.

                *Parameter example:*
                  ``domain_axis='domainaxis'``

            domain_topology: `str`, optional
                The base name for keys of domain topology constructs.

                *Parameter example:*
                  ``'domaintopology'``

                .. versionadded:: (cfdm) 1.11.0.0

            cell_connectivity: `str`, optional
                The base name for keys of cell connectivity
                constructs.

                *Parameter example:*
                  ``'cellconnectivity'``

                .. versionadded:: (cfdm) 1.11.0.0

            cell_method: `str`, optional
                The base name for keys of cell method constructs.

                *Parameter example:*
                  ``cell_method='cellmethod'``

            {{init source: optional}}

            copy: `bool`, optional
                If True (the default) then deep copy metadata constructs
                from those of *source* prior to initialisation, else they
                are not deep copied.

            _ignore: sequence of `str`, optional
                Ignores the given construct types.

                *Parameter example:*
                  ``_ignore=('cell_method', 'field_ancillary')``

        """
        _ignore = tuple(_ignore)

        if source is not None:
            if _view:
                self.__dict__ = source.__dict__.copy()
                self._ignore = _ignore
                return

            source_constructs = source._constructs

            self._field_data_axes = source._field_data_axes

            self._key_base = source._key_base.copy()
            self._array_constructs = source._array_constructs.copy()
            self._non_array_constructs = source._non_array_constructs.copy()
            self._construct_axes = source._construct_axes.copy()
            self._construct_type = source._construct_type.copy()

            self_construct_axes = self._construct_axes
            self_construct_type = self._construct_type

            d = {}
            for construct_type in source._array_constructs:
                if construct_type in _ignore:
                    for cid in source_constructs.get(construct_type, ()):
                        self_construct_axes.pop(cid, None)
                        self_construct_type.pop(cid, None)

                    continue

                if construct_type not in source_constructs:
                    continue

                if copy:
                    new_v = {}
                    for cid, construct in source_constructs[
                        construct_type
                    ].items():
                        new_v[cid] = construct.copy(data=_use_data)
                else:
                    new_v = source_constructs[construct_type].copy()

                d[construct_type] = new_v

            for construct_type in source._non_array_constructs:
                if construct_type in _ignore:
                    for cid in source_constructs.get(construct_type, ()):
                        self_construct_type.pop(cid, None)

                    continue

                if construct_type not in source_constructs:
                    continue

                if copy:
                    new_v = {}
                    for cid, construct in source_constructs[
                        construct_type
                    ].items():
                        new_v[cid] = construct.copy()
                else:
                    new_v = source_constructs[construct_type].copy()

                d[construct_type] = new_v

            self._constructs = d

            self._ignore = ()

            return

        self._ignore = _ignore

        self._field_data_axes = None

        self._key_base = {}

        self._array_constructs = set()
        self._non_array_constructs = set()

        self._construct_axes = {}

        # The construct type for each key. For example:
        # {'domainaxis1':'domain_axis',
        #  'auxiliarycoordinate3':'auxiliary_coordinate'}
        self._construct_type = {}

        self._constructs = {}

        if auxiliary_coordinate:
            self._key_base["auxiliary_coordinate"] = auxiliary_coordinate
            self._array_constructs.add("auxiliary_coordinate")

        if dimension_coordinate:
            self._key_base["dimension_coordinate"] = dimension_coordinate
            self._array_constructs.add("dimension_coordinate")

        if domain_ancillary:
            self._key_base["domain_ancillary"] = domain_ancillary
            self._array_constructs.add("domain_ancillary")

        if field_ancillary:
            self._key_base["field_ancillary"] = field_ancillary
            self._array_constructs.add("field_ancillary")

        if cell_measure:
            self._key_base["cell_measure"] = cell_measure
            self._array_constructs.add("cell_measure")

        if domain_topology:
            self._key_base["domain_topology"] = domain_topology
            self._array_constructs.add("domain_topology")

        if cell_connectivity:
            self._key_base["cell_connectivity"] = cell_connectivity
            self._array_constructs.add("cell_connectivity")

        if domain_axis:
            self._key_base["domain_axis"] = domain_axis
            self._non_array_constructs.add("domain_axis")

        if coordinate_reference:
            self._key_base["coordinate_reference"] = coordinate_reference
            self._non_array_constructs.add("coordinate_reference")

        if cell_method:
            self._key_base["cell_method"] = cell_method
            self._non_array_constructs.add("cell_method")

        for x in self._array_constructs:
            self._constructs[x] = {}

        for x in self._non_array_constructs:
            self._constructs[x] = {}

    def __contains__(self, key):
        """Implements membership test operators for construct keys.

        x.__contains__(y) <==> y in x

        .. versionadded:: (cfdm) 1.7.0

        """
        return self.construct_type(key) is not None

    def __copy__(self):
        """Called by the `copy.copy` standard library function.

        .. versionadded:: (cfdm) 1.7.0

        """
        return self.shallow_copy()

    def __deepcopy__(self, memo):
        """Called by the `copy.deepcopy` standard library function.

        .. versionadded:: (cfdm) 1.7.0

        """
        return self.copy()

    def __getitem__(self, key):
        """Return a construct with the given key.

        x.__getitem__(y) <==> x[y]

        .. versionadded:: (cfdm) 1.7.0

        """
        construct_type = self.construct_type(key)
        if construct_type is None:
            raise KeyError(key)

        return self._constructs[construct_type][key]

    def __iter__(self):
        """Called when an iterator is required.

        x.__iter__() <==> iter(x)

        .. versionadded:: (cfdm) 1.7.0

        """
        return iter(self.todict())

    def __len__(self):
        """Return the number of constructs.

        x.__len__() <==> len(x)

        .. versionadded:: (cfdm) 1.7.0

        """
        n = len(self._construct_type)
        ignore = self._ignore
        if ignore:
            constructs = self._constructs
            for ctype in ignore:
                n -= len(constructs[ctype])

        return n

    def _construct_dict(self, construct_type, copy=False):
        """Return the dictionary of constructs of a particular type.

        .. versionadded:: (cfdm) 1.8.9.0

        :Parameters:

            copy: `bool`, optional
                If True and some constructs exisit then the dictionary
                is a shallow copy of that stored in the `_constructs`
                attribute, otherwise it is the same dictionary.

        :Returns:

            `dict`
                The dictionary, keyed by construct identifiers with
                construct values.

        """
        if construct_type in self._ignore:
            return {}

        out = self._constructs.get(construct_type)
        if out is None:
            return {}

        if copy:
            out = out.copy()

        return out

    def _del_data_axes(self, k, *d):
        """Remove and return a construct's axes, if any.

        If k is not found, d is returned if given, otherwise KeyError is
        raised

        """
        return self._construct_axes.pop(k, *d)

    def _view(self, ignore=()):
        """Returns a new container with a view of the same constructs.

        A new `{{class}}` instance is returned that contains the same
        metadata construct instances

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            ignore: sequence of `str`, optional
                Return a view that ignores the given construct types.

                *Parameter example:*
                  ``ignore=('cell_method', 'field_ancillary')``

        :Returns:

            `{{class}}`
                The new constructs container.

        """
        return type(self)(source=self, _ignore=ignore, _view=True)

    def _check_construct_type(self, construct_type, default=ValueError()):
        """Check the type of a metadata construct is valid.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            construct_type: `str`
                The construct type to be checked.

            default: `bool`, optional
                Return the value of the *default* parameter if
                construct type is not valid.

                {{default Exception}}

        :Returns:

            `str` or `None`
                Return the type of of the construct, or if the input
                construct was given as `None`, `None` is returned.

        """
        if (
            construct_type not in self._ignore
            and construct_type in self._key_base
        ) or construct_type is None:
            return construct_type

        if default is None:
            return default

        return self._default(
            default, f"Invalid construct type {construct_type!r}"
        )

    def _construct_type_description(self, construct_type):
        """Format the description of the type of a metadata construct.

        Type name components are formatted to be whitespace-delimited to
        effective words for the purposes of printing to the user.

        """
        return construct_type.replace("_", " ")

    def _del_construct(self, key, default=ValueError()):
        """Remove a metadata construct.

        If a domain axis construct is selected for removal then it
        can't be spanned by any metadata construct data arrays, nor be
        referenced by any cell method constructs.

        However, a domain ancillary construct may be removed even if
        it is referenced by coordinate reference construct. In this
        case the reference is replace with `None`.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `_get_construct`, `_set_construct`

        :Parameters:

            key: `str`
                The key of the construct to be removed.

                *Parameter example:*
                  ``key='auxiliarycoordinate0'``

            default: optional
                Return the value of the *default* parameter if the
                construct can not be removed, or does not exist.

                {{default Exception}}

        :Returns:

                The removed construct.

        **Examples**

        >>> f = {{package}}.example_field(3)
        >>> c = f.constructs
        >>> x = c._del_construct('auxiliarycoordinate2')

        """
        data_axes = self.data_axes()

        domain_axes = self._construct_dict("domain_axis")

        if key in domain_axes:
            # Fail if the domain axis construct is spanned by a data
            # array
            for xid, axes in data_axes.items():
                if key in axes:
                    if default is None:
                        return default

                    raise ValueError(
                        f"Can't remove domain axis construct {key!r} that "
                        f"spans the data array of metadata construct {xid!r}"
                    )

            cell_methods = self._construct_dict("cell_method")
            for xid, cm in cell_methods.items():
                if key in cm.get_axes(()):
                    if default is None:
                        return default

                    raise ValueError(
                        f"Can't remove domain axis construct {key!r} "
                        "that is referenced by cell method construct "
                        f"{xid!r}"
                    )
        else:
            # Remove references to the removed construct in coordinate
            # reference constructs
            coordinate_references = self._construct_dict(
                "coordinate_reference"
            )
            for ref in coordinate_references.values():
                coordinate_conversion = ref.coordinate_conversion
                for (
                    term,
                    value,
                ) in coordinate_conversion.domain_ancillaries().items():
                    if key == value:
                        coordinate_conversion.set_domain_ancillary(term, None)

                ref.del_coordinate(key, None)

        out = self._pop(key, None)

        if out is None:
            if default is None:
                return default

            return self._default(
                default, f"Can't remove non-existent construct {key!r}"
            )

        return out

    def _dictionary(self, copy=False):
        """Deprecated at 1.8.9.0.

        Use `todict` instead.

        """
        return self.todict()  # pragma: no cover

    def _set_construct(self, construct, key=None, axes=None, copy=True):
        """Set a metadata construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `_del_construct`, `_get_construct`,
                     `_set_construct_data_axes`

        :Parameters:

            construct:
                The metadata construct to be inserted.

            key: `str`, optional
                The construct identifier to be used for the
                construct. If not set then a new, unique identifier is
                created automatically. If the identifier already
                exists then the existing construct will be replaced.

                *Parameter example:*
                  ``key='cellmeasure0'``

            axes: (sequence of) `str`, optional
                The construct identifiers of the domain axis
                constructs spanned by the data array. An exception is
                raised if used for a metadata construct that can not
                have a data array, i.e. domain axis, cell method and
                coordinate reference constructs.

                The axes may also be set afterwards with the
                `_set_construct_data_axes` method.

                *Parameter example:*
                  ``axes='domainaxis1'``

                *Parameter example:*
                  ``axes=['domainaxis1']``

                *Parameter example:*
                  ``axes=('domainaxis1', 'domainaxis0')``

            copy: `bool`, optional
                If True then return a copy of the unique selected
                construct. By default the construct is copied.

        :Returns:

             `str`
                The construct identifier for the construct.

        **Examples**

        >>> f = {{package}}.example_field(3)
        >>> c = f.constructs
        >>> c
        <{{repr}}Constructs: auxiliary_coordinate(6), domain_axis(2)>
        >>> d = {{package}}.DomainAxis(10)
        >>> key1 = c._set_construct(d, copy=False)
        >>> key1
        'domainaxis2'
        >>> print(c.filter_by_type('domain_axis'))
        Constructs:
        {'domainaxis0': <{{repr}}DomainAxis: size(4)>,
         'domainaxis1': <{{repr}}DomainAxis: size(9)>,
         'domainaxis2': <{{repr}}DomainAxis: size(10)>}

        >>> a = {{package}}.CellMeasure(measure='area', properties={'units': 'm2'})
        >>> key2 = c._set_construct(a, key='my_area_cell_measure')
        >>> key2
        'my_area_cell_measure'
        >>> v = {{package}}.CellMeasure(measure='volume', properties={'units': 'm3'})
        >>> key3 = c._set_construct(
        ...     v, key='my_volume_cell_measure', axes='domainaxis2')
        >>> key3
        'my_volume_cell_measure'
        >>> c
        <{{repr}}Constructs: auxiliary_coordinate(6), cell_measure(2), domain_axis(3)>
        >>> print(c.filter_by_type('cell_measure'))
        Constructs:
        {'my_area_cell_measure': <{{repr}}CellMeasure: measure:area m2>,
         'my_volume_cell_measure': <{{repr}}CellMeasure: measure:volume m3>}

        """
        construct_type = self._check_construct_type(construct.construct_type)

        if key is None:
            # Create a new construct identifier
            key = self.new_identifier(construct_type)

        if construct_type in self._array_constructs:
            # ---------------------------------------------------------
            # The construct could have a data array
            # ---------------------------------------------------------
            if axes is not None:
                self._set_construct_data_axes(
                    key=key, axes=axes, construct=construct
                )
        elif axes is not None:
            raise ValueError(
                f"Can't set {construct!r}: Can't provide domain axis "
                "constructs for "
                f"{self._construct_type_description(construct_type)} "
                "construct"
            )

        # Record the construct type
        self._construct_type[key] = construct_type

        if copy:
            # Create a deep copy of the construct
            construct = construct.copy()

        # Insert the construct
        self._constructs[construct_type][key] = construct

        # Return the identifier of the construct
        return key

    def _set_construct_data_axes(self, key, axes, construct=None):
        """Set domain axis constructs for construct identifiers.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            key: `str`, optional
                The construct identifier of metadata construct.

                *Parameter example:*
                  ``key='cellmeasure0'``

            axes: (sequence of) `str`
                The construct identifiers of the domain axis
                constructs spanned by the data array. An exception is
                raised if used for a metadata construct that can not
                have a data array, such as a domain axis construct.

                *Parameter example:*
                  ``axes='domainaxis1'``

                *Parameter example:*
                  ``axes=['domainaxis1']``

                *Parameter example:*
                  ``axes=['domainaxis1', 'domainaxis0']``

        :Returns:

            `None`

        **Examples**

        >>> f = {{package}}.example_field(3)
        >>> c = f.constructs
        >>> c
        <{{repr}}Constructs: auxiliary_coordinate(6), domain_axis(2)>
        >>> a = {{package}}.CellMeasure(measure='area', properties={'units': 'm2'})
        >>> area_measure_key = c.set_construct(a)
        >>> f._set_construct_data_axes(area_measure_key, axes='domainaxis0')

        """
        if construct is None:
            if self.construct_type(key) is None:
                raise ValueError(
                    "Can't set axes for non-existent construct "
                    f"identifier {key!r}"
                )

            construct = self[key]

        if isinstance(axes, str):
            axes = (axes,)

        domain_axes = self._construct_dict("domain_axis")

        axes_shape = []
        for axis in axes:
            if axis not in domain_axes:
                raise ValueError(
                    f"Can't set {construct!r} domain axes: "
                    f"Domain axis {axis!r} does not exist"
                )

            axes_shape.append(domain_axes[axis].get_size())

        axes_shape = tuple(axes_shape)
        try:
            # Note: The construct shape might be different from the
            #       construct's data shape. For instance, this will be
            #       the case for an auxiliary coordinate construct
            #       that has bounds but no cell coordinates; or for
            #       any domain topology construct or cell connectivity
            #       construct.
            if construct.shape != axes_shape:
                raise ValueError(
                    f"Can't set {construct!r}: Shape of {construct.shape!r} "
                    "does not match the shape required by domain axes "
                    f"{tuple(axes)}: {axes_shape}"
                )
        except AttributeError:
            pass

        self._construct_axes[key] = tuple(axes)

    # ----------------------------------------------------------------
    # Private dictionary-like methods
    # ----------------------------------------------------------------
    def _clear(self):
        """D.clear() -> None.

        Remove all items from D.

        """
        self._ignore = ()
        self._key_base.clear()
        self._array_constructs.clear()
        self._non_array_constructs.clear()
        self._construct_axes.clear()
        self._construct_type.clear()
        self._constructs.clear()

    def _pop(self, k, *d):
        """Removes specified key and returns the corresponding value.

        D.pop(k[,d]) -> v, remove specified key and return the
        corresponding value. If key is not found, d is returned if
        given, otherwise KeyError is raised.

        """
        construct_type = self.construct_type(k)
        if construct_type is not None:
            self._del_data_axes(k, None)
            del self._construct_type[k]
            return self._constructs[construct_type].pop(k)

        if d:
            return d[0]

        raise KeyError(k)

    def _update(self, other):
        """D.update(E) -> None.

        Update D from E.

        """
        self._ignore = tuple(set(self._ignore).union(other._ignore))

        self._key_base.update(other._key_base)
        self._array_constructs.update(other._array_constructs)
        self._non_array_constructs.update(other._non_array_constructs)
        self._construct_axes.update(other._construct_axes)
        self._construct_type.update(other._construct_type)
        self._constructs.update(other._constructs)

    # ----------------------------------------------------------------
    # Dictionary-like methods
    # ----------------------------------------------------------------
    def get(self, key, default=None):
        """Returns the construct for the key, else default.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `items`, `keys`, `values`

        """
        construct_type = self.construct_type(key)
        if construct_type is not None:
            return self._constructs[construct_type][key]

        return default

    def items(self):  # SB NOTE: flaky doctest due to dict_items order
        """Return the items as (construct key, construct) pairs.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get`, `keys`, `values`

        :Returns:

            `dict_items`
                The construct key and constructs respectively as
                key-value pairs in a Python `dict_items` iterator.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> c = f.constructs
        >>> c_items = c.items()
        >>> print(c_items)
        dict_items([('dimensioncoordinate0', <{{repr}}DimensionCoordinate: latitude(5) degrees_north>),
                    ('dimensioncoordinate1', <{{repr}}DimensionCoordinate: longitude(8) degrees_east>),
                    ('dimensioncoordinate2', <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >),
                    ('domainaxis0', <{{repr}}DomainAxis: size(5)>),
                    ('domainaxis1', <{{repr}}DomainAxis: size(8)>),
                    ('domainaxis2', <{{repr}}DomainAxis: size(1)>),
                    ('cellmethod0', <{{repr}}CellMethod: area: mean>)])
        >>> type(c_items)
        <class 'dict_items'>
        >>> dict(c_items)
        {'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >,
         'cellmethod0': <{{repr}}CellMethod: area: mean>,
         'domainaxis0': <{{repr}}DomainAxis: size(5)>,
         'domainaxis1': <{{repr}}DomainAxis: size(8)>,
         'domainaxis2': <{{repr}}DomainAxis: size(1)>}

        """
        return self.todict().items()

    def keys(self):  # SB NOTE: flaky doctest due to dict_items order
        """Return all of the construct keys, in arbitrary order.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get`, `items`, `values`

        :Returns:

            `dict_keys`
                The construct keys as a Python `dict_keys` iterator.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> c = f.constructs
        >>> c_keys = c.keys()
        >>> print(c_keys)
        dict_keys(['domainaxis0',
                   'domainaxis1',
                   'domainaxis2',
                   'dimensioncoordinate0',
                   'dimensioncoordinate1',
                   'dimensioncoordinate2',
                   'cellmethod0'])
        >>> type(c_keys)
        <class 'dict_keys'>
        >>> list(c_keys)
        ['domainaxis0',
         'domainaxis1',
         'domainaxis2',
         'dimensioncoordinate0',
         'dimensioncoordinate1',
         'dimensioncoordinate2',
         'cellmethod0']

        """
        return self.todict().keys()

    def values(self):  # SB NOTE: flaky doctest due to dict_items order
        """Return all of the metadata constructs, in arbitrary order.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get`, `items`, `keys`

        :Returns:

            `dict_values`
                The constructs as a Python `dict_values` iterator.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> c = f.constructs
        >>> c_values = c.values()
        >>> print(c_values)
        dict_values([<{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
                     <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
                     <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >,
                     <{{repr}}DomainAxis: size(5)>,
                     <{{repr}}DomainAxis: size(8)>,
                     <{{repr}}DomainAxis: size(1)>,
                     <{{repr}}CellMethod: area: mean>])
        >>> type(c_values)
        <class 'dict_values'>
        >>> list(c_values)
        [<{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >,
         <{{repr}}DomainAxis: size(5)>,
         <{{repr}}DomainAxis: size(8)>,
         <{{repr}}DomainAxis: size(1)>,
         <{{repr}}CellMethod: area: mean>]

        """
        return self.todict().values()

    def construct_type(self, key):
        """Return the type of a construct for a given key.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `construct_types`

        :Parameters:

            key: `str`
                A construct identifier.

        :Returns:

            `str` or `None`
                The construct type, or `None` if the *key* is not
                present.

        """
        ctype = self._construct_type.get(key)
        if ctype is not None and ctype in self._ignore:
            return

        return ctype

    def construct_types(self):
        """Return all of the construct types for all keys.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `construct_type`

        """
        out = self._construct_type.copy()

        ignore = self._ignore
        if ignore:
            for key, ctype in self._construct_type.items():
                if ctype in ignore:
                    del out[key]

        return out

    def copy(self, data=True):
        """Return a deep copy.

        ``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            {{data: `bool`, optional}}

        :Returns:

            `{{class}}`
                The deep copy.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> c = f.constructs
        >>> k = c.copy()
        >>> k = c.copy(data=False)

        """
        return type(self)(
            source=self,
            copy=True,
            _ignore=self._ignore,
            _view=False,
            _use_data=data,
        )

    def data_axes(self):
        """Returns the axes spanned by the data.

        Specifically, returns the domain axis constructs spanned by
        metadata construct data.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `dict`
                The keys of the domain axes constructs spanned by
                metadata construct data.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> c = f.constructs
        >>> print(c)
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: area: mean>,
         'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >,
         'domainaxis0': <{{repr}}DomainAxis: size(5)>,
         'domainaxis1': <{{repr}}DomainAxis: size(8)>,
         'domainaxis2': <{{repr}}DomainAxis: size(1)>}
        >>> c.data_axes()
        {'dimensioncoordinate0': ('domainaxis0',),
         'dimensioncoordinate1': ('domainaxis1',),
         'dimensioncoordinate2': ('domainaxis2',)}

        """
        out = self._construct_axes.copy()

        ignore = self._ignore
        if not ignore:
            return out

        non_data_constructs = self._non_array_constructs
        for construct_type, keys in self._constructs.items():
            if construct_type in non_data_constructs:
                continue

            if construct_type in ignore:
                for key in keys:
                    out.pop(key, None)

        return out

    def filter_by_type(self, *types, todict=True, cached=None):
        """Select metadata constructs by type.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            types: optional
                Select constructs that have are of any of the given
                types.

                A type is specified by one of the following strings:

                ==========================  ==========================
                *type*                      Construct selected
                ==========================  ==========================
                ``'domain_ancillary'``      Domain ancillary
                ``'dimension_coordinate'``  Dimension coordinate
                ``'domain_axis'``           Domain axis
                ``'auxiliary_coordinate'``  Auxiliary coordinate
                ``'cell_measure'``          Cell measure
                ``'coordinate_reference'``  Coordinate reference
                ``'domain_topology'``       Domain topology
                ``'cell_connectivity'``     Cell connectivity
                ``'cell_method'``           Cell method
                ``'field_ancillary'``       Field ancillary
                ==========================  ==========================

                If no types are provided then all constructs are
                selected.

        :Parameters:

            {{todict: `bool`, optional}}

            {{cached: optional}}

        :Returns:

            `{{class}}` or `dict`
                The selected constructs and their construct keys.

        **Examples**

        Select dimension coordinate constructs:

        >>> f = {{package}}.example_field(1)
        >>> c = f.constructs
        >>> d = c.filter_by_type('dimension_coordinate')
        >>> d
        <{{repr}}Constructs: dimension_coordinate(4)>

        Select dimension coordinate and field ancillary constructs:

        >>> k = c.filter_by_type(
        ...     'dimension_coordinate', 'field_ancillary')
        >>> k
        <{{repr}}Constructs: dimension_coordinate(4), field_ancillary(1)>

        """
        if cached is not None:
            return cached

        if todict:
            # Return a dictionary
            if not types:
                return self.todict()

            out = self._construct_dict(types[0], copy=True)

            if len(types) > 1:
                ignore = self._ignore
                constructs = self._constructs
                for t in types[1:]:
                    if t in constructs and t not in ignore:
                        out.update(constructs[t])

            return out

        # Still here? Then return a `Constructs` object
        if types:
            # Ignore the all but the requested construct types
            ignore = set(self._key_base)
            ignore.difference_update(types)
            ignore.update(self._ignore)
        else:
            # Keep all construct types
            ignore = self._ignore

        return self.shallow_copy(_ignore=ignore)

    def get_data_axes(self, key, default=ValueError()):
        """Return the keys of the axes spanned by a construct's data.

        Specifically, returns the keys of the domain axis constructs
        spanned by the data of a metadata construct.

        .. versionadded:: (cfdm) 1.8.9.0

        :Parameters:

            key: `str`
                Specify a metadata construct.

                *Parameter example:*
                  ``key='auxiliarycoordinate0'``

            default: optional
                Return the value of the *default* parameter if the data
                axes have not been set.

                {{default Exception}}

        :Returns:

            `tuple`
                The keys of the domain axis constructs spanned by the data.

        """
        data_axes = self._construct_axes.get(key)
        if data_axes is not None:
            ignore = self._ignore
            if not ignore or self._construct_type[key] not in ignore:
                return data_axes

        if default is None:
            return default

        return self._default(default, f"No data axes for construct {key!r}")

    def key(self, default=ValueError()):
        """Return the construct key of the sole metadata construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get`, `keys`, `value`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if there
                is not exactly one construct.

                {{default Exception}}

        :Returns:

            `str`
                The construct key.

        **Examples**

        >>> f = {{package}}.Field()
        >>> c = f.constructs
        >>> d = {{package}}.DimensionCoordinate(
        ...     properties={
        ...         'standard_name': 'latitude', 'units': 'degrees_north'},
        ...     data={{package}}.Data(range(5))
        ... )
        >>> c._set_construct(d)
        'dimensioncoordinate0'
        >>> print(c)
        Constructs:
        {'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>}
        >>> c.key()
        'dimensioncoordinate0'
        >>> c.value()
        <{{repr}}DimensionCoordinate: latitude(5) degrees_north>

        """
        d = self.todict()
        n = len(d)
        if n == 1:
            key, _ = d.popitem()
            return key

        if default is None:
            return default

        if not n:
            return self._default(default, "Can't get key for zero constructs")

        return self._default(
            default, f"Can't get key for more than one ({n}) constructs"
        )

    def new_identifier(self, construct_type):
        """Return a new, unused construct key.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            construct_type: `str`
                The construct type for which the identifier is being
                created.

                *Parameter example:*
                  ``construct_type='dimension_coordinate'``

        :Returns:

            `str`
                The new construct identifier.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> c = f.constructs
        >>> c.keys()
        ['domainaxis0',
         'domainaxis1',
         'domainaxis2',
         'dimensioncoordinate0',
         'dimensioncoordinate1',
         'dimensioncoordinate2',
         'cellmethod0']
        >>> c.new_identifier('domain_axis')
        'domainaxis3'
        >>> c.new_identifier('cell_method')
        'cellmethod1'

        """
        construct_type = self._check_construct_type(construct_type)

        keys = self._constructs[construct_type]

        n = len(keys)
        key_base = self._key_base[construct_type]
        key = f"{key_base}{n}"
        while key in keys:
            n += 1
            key = f"{key_base}{n}"

        return key

    def ordered(self):
        """Return the constructs in their predetermined order.

        For cell method constructs, the predetermined order is that in
        which they where added. There is no predetermined ordering for
        all other construct types, and a exception is raised if any
        non-cell method constructs are present.

        Deprecated at version (cfdm) 1.9.0.0, since all metadata
        constructs are now stored in the order in which they where
        added.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

             `collections.OrderedDict`
                 The constructs and their construct keys, in their
                 predetermined order.

        **Examples**

        >>> f = {{package}}.Field()
        >>> c = f.constructs
        >>> c.ordered()
        {}
        >>> m1 = {{package}}.CellMethod(axes=['domainaxis1'], method='mean')
        >>> c._set_construct(m1, key='cell_method_for_axis_1')
        'cell_method_for_axis_1'
        >>> m2 = {{package}}.CellMethod(axes=['domainaxis0'], method='minimum')
        >>> c._set_construct(m2, key='cell_method_for_axis_0')
        'cell_method_for_axis_0'
        >>> print(c)
        Constructs:
        {'cell_method_for_axis_0': <{{repr}}CellMethod: domainaxis0: minimum>,
         'cell_method_for_axis_1': <{{repr}}CellMethod: domainaxis1: mean>}
        >>> c.ordered()
        OrderedDict([('cell_method_for_axis_1', <{{repr}}CellMethod: domainaxis1: mean>),
                     ('cell_method_for_axis_0', <{{repr}}CellMethod: domainaxis0: minimum>)])

        """
        raise DeprecationWarning(
            "This method is no longer required from v1.9.0.0, and"
            "will be removed at a later date"
        )

    def replace(self, key, construct, axes=None, copy=True):
        """Replace one metadata construct with another.

        .. note:: No checks on the axes are done.

        """
        construct_type = self.construct_types().get(key)
        if construct_type is None:
            raise ValueError(f"Can't replace non-existent construct {key!r}")

        if axes is not None and construct_type in self._array_constructs:
            self._construct_axes[key] = tuple(axes)

        if copy:
            construct = construct.copy()

        self._constructs[construct_type][key] = construct

    def shallow_copy(self, _ignore=None):
        """Return a shallow copy.

        ``copy.copy(f)`` is equivalent to ``f.shallow_copy()``.

        Any in-place changes to the actual constructs of the copy will
        not be seen in the original `{{class}}` object, and vice
        versa, but the copy and filter history are otherwise
        independent.

        .. versionadded:: (cfdm) 1.8.9.0

        .. seealso:: `view`

        :Returns:

            `{{class}}`
                The shallow copy.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> c = f.constructs
        >>> k = c.shallow_copy()

        """
        if _ignore is None:
            _ignore = self._ignore

        return type(self)(
            source=self, copy=False, _ignore=_ignore, _view=False
        )

    def todict(self, copy=False):
        """Return a dictionary of the metadata constructs.

        .. versionadded:: (cfdm) 1.8.9.0

        :Parameters:

            copy: `bool`, optional
                If True then deep copy the metadata construct values.

        :Returns:

            `dict`
                The dictionary representation, keyed by construct
                identifiers with metadata construct values.

        """
        out = {}
        ignore = self._ignore
        for key, value in self._constructs.items():
            if key not in ignore:
                out.update(value)

        if copy:
            for key, construct in out.items():
                out[key] = construct.copy()

        return out

    def value(self, default=ValueError()):
        """Return the sole metadata construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get`, `key`, `values`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if there
                is not exactly one construct.

                {{default Exception}}

        :Returns:

                The metadata construct.

        **Examples**

        >>> f = {{package}}.Field()
        >>> c = f.constructs
        >>> d = {{package}}.DimensionCoordinate(
        ...     properties={
        ...         'standard_name': 'latitude', 'units': 'degrees_north'},
        ...     data={{package}}.Data(range(5))
        ... )
        >>> c._set_construct(d)
        'dimensioncoordinate0'
        >>> print(c)
        Constructs:
        {'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>}
        >>> c.key()
        'dimensioncoordinate0'
        >>> c.value()
        <{{repr}}DimensionCoordinate: latitude(5) degrees_north>

        """
        d = self.todict()
        n = len(d)
        if n == 1:
            _, construct = d.popitem()
            return construct

        if default is None:
            return default

        if not d:
            return self._default(default, "Can't return zero constructs")

        return self._default(
            default, f"Can't return more than one ({n}) constructs"
        )



class LinearInterpolation:
    """Mixin class for linear interpolation.

    See CF appendix J "Coordinate Interpolation Methods".

    .. versionadded:: (cfdm) 1.10.0.0

    """

    def _linear_interpolation(self, ua, ub, d1, s=None):
        """Interpolate linearly between two points.

        u = fl(ua, ub, s) = ua + s*(ub-ua)
                          = ua*(1-s) + ub*s

        See CF appendix J "Coordinate Interpolation Methods".

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `_s`

        :Parameters:

            ua, ub: `numpy.ndarray`
                The tie points at locations A and B, in the sense of
                CF appendix J Figure J.1.

            {{d1: `int`}}

            {{s: array_like, optional}}

        :Returns:

            `numpy.ndarray`
                The result of interpolating the tie points to interior
                locations implied by *s*.

        """
        s = self._s(d1, s=s)
        u = ua + s * (ub - ua)
        return u



from .fragmentfilearray import FragmentFileArray
from .fragmenth5netcdfarray import FragmentH5netcdfArray
from .fragmentnetcdf4array import FragmentNetCDF4Array
from .fragmentuniquevaluearray import FragmentUniqueValueArray



import os
import platform
import sys
from pickle import dumps, loads

from . import __cf_version__, __file__, __version__


def environment(display=True, paths=True):
    """Return the names, versions and paths of all dependencies.

    .. versionadded:: (cfdm) 1.7.0

    :Parameters:

        display: `bool`, optional
            If True (the default) then display the description of the
            environment as a string. If False the description is
            instead returned as a list.

        paths: `bool`, optional
            If True (the default) then output the locations of each
            package. If False the locations are not included.

    :Returns:

        `None` or `list`
            If *display* is True then the description of the
            environment is printed and `None` is returned. Otherwise
            the description is returned as in a `list`.

    **Examples**

    >>> cfdm.core.environment(paths=False)
    Platform: Linux-5.15.0-92-generic-x86_64-with-glibc2.35
    Python: 3.11.4
    packaging: 23.0
    numpy: 1.25.2
    cfdm.core: 1.11.2.0

    """
    import numpy as np
    import packaging

    dependency_version_paths_mapping = {
        "Platform": (platform.platform(), ""),
        "Python": (platform.python_version(), sys.executable),
        "packaging": (
            packaging.__version__,
            os.path.abspath(packaging.__file__),
        ),
        "numpy": (np.__version__, os.path.abspath(np.__file__)),
        "cfdm.core": (__version__, os.path.abspath(__file__)),
    }
    string = "{0}: {1!s}"
    if paths:  # include path information, else exclude, when unpacking tuple
        string += " {2!s}"
    out = [
        string.format(dep, *info)
        for dep, info in dependency_version_paths_mapping.items()
    ]

    if display:
        print("\n".join(out))  # pragma: no cover
    else:
        return out


def CF():
    """The version of the CF conventions.

    This indicates which version of the CF conventions are represented
    by this release of the cfdm.core package, and therefore the
    version can not be changed.

    .. versionadded:: (cfdm) 1.7.0

    :Returns:

        `str`
            The version of the CF conventions represented by this
            release of the cfdm.core package.

    **Examples**

    >>> cfdm.core.CF()
    '1.10'

    """
    return __cf_version__


def deepcopy(x):
    """Use pickle/unpickle to deep copy an object.

    For the types of inputs expected, pickle/unpickle should a) work and
    b) be "not slower, sometimes much faster" than `copy.deepcopy`.

    """
    return loads(dumps(x))



from .bilinearinterpolation import BiLinearInterpolation
from .biquadraticlatitudelongitudeinterpolation import (
    BiQuadraticLatitudeLongitudeInterpolation,
)
from .linearinterpolation import LinearInterpolation
from .pointtopology import PointTopology
from .quadraticgeographicinterpolation import QuadraticGeographicInterpolation
from .quadraticinterpolation import QuadraticInterpolation
from .quadraticlatitudelongitudeinterpolation import (
    QuadraticLatitudeLongitudeInterpolation,
)
